# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import math
import os
import os.path as osp
import time
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

from nncf.common.hardware.config import HWConfigType
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.common.utils.debug import is_debug
from nncf.common.utils.os import safe_open
from nncf.config.schemata.defaults import AUTOQ_EVAL_SUBSET_RATIO
from nncf.config.schemata.defaults import AUTOQ_ITER_NUMBER
from nncf.config.schemata.defaults import AUTOQ_WARMUP_ITER_NUMBER
from nncf.config.schemata.defaults import PRECISION_INIT_BITWIDTHS
from nncf.torch.automl.agent.ddpg.ddpg import DDPG
from nncf.torch.quantization.precision_constraints import HardwareQuantizationConstraints
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitializer
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitParams
from nncf.torch.structures import AutoQPrecisionInitArgs


class AutoQPrecisionInitParams(BasePrecisionInitParams):
    def __init__(
        self,
        user_init_args: AutoQPrecisionInitArgs,
        dump_autoq_data: bool = False,
        iter_number: int = 0,
        warmup_iter_number: int = None,
        compression_ratio: float = None,
        eval_subset_ratio: float = None,
        ddpg_hparams_dict: Dict = None,
        hw_cfg_type: HWConfigType = None,
        skip_constraint: bool = False,
        finetune: bool = False,
        bits: List[int] = None,
    ):
        super().__init__(user_init_args)
        self.dump_autoq_data = dump_autoq_data
        self.iter_number = iter_number
        self.compression_ratio = compression_ratio
        self.eval_subset_ratio = eval_subset_ratio
        self.warmup_iter_number = warmup_iter_number
        if ddpg_hparams_dict is None:
            self.ddpg_hparams_dict = {}
        else:
            self.ddpg_hparams_dict = ddpg_hparams_dict
        self.hw_cfg_type = hw_cfg_type
        self.skip_constraint = skip_constraint
        self.finetune = finetune
        self.bits = bits

    @classmethod
    def from_config(
        cls,
        autoq_init_config_dict: Dict,
        user_init_args: AutoQPrecisionInitArgs,
        target_hw_config_type: Optional[HWConfigType],
    ) -> "AutoQPrecisionInitParams":
        dict_copy = autoq_init_config_dict.copy()
        dump_autoq_data = dict_copy.pop("dump_init_precision_data", False)
        iter_number = dict_copy.pop("iter_number", AUTOQ_ITER_NUMBER)
        compression_ratio = dict_copy.pop("compression_ratio", 0.15)
        eval_subset_ratio = dict_copy.pop("eval_subset_ratio", AUTOQ_EVAL_SUBSET_RATIO)
        warmup_iter_number = dict_copy.pop("warmup_iter_number", AUTOQ_WARMUP_ITER_NUMBER)
        skip_constraint = dict_copy.pop("skip_constraint", False)
        finetune = dict_copy.pop("finetune", False)
        bits = dict_copy.pop("bits", PRECISION_INIT_BITWIDTHS)

        return cls(
            user_init_args=user_init_args,
            dump_autoq_data=dump_autoq_data,
            iter_number=iter_number,
            warmup_iter_number=warmup_iter_number,
            ddpg_hparams_dict=dict_copy,
            hw_cfg_type=target_hw_config_type,
            compression_ratio=compression_ratio,
            eval_subset_ratio=eval_subset_ratio,
            skip_constraint=skip_constraint,
            finetune=finetune,
            bits=bits,
        )


class AutoQPrecisionInitializer(BasePrecisionInitializer):
    def __init__(
        self,
        algo: "ExperimentalQuantizationController",  # noqa: F821
        params: AutoQPrecisionInitParams,
        hw_precision_constraints: HardwareQuantizationConstraints,
    ):
        super().__init__(algo, params, hw_precision_constraints)
        self.quantization_controller = algo
        self._params = params
        self._init_args = params.user_init_args
        self._dump_autoq_data = params.dump_autoq_data
        self._iter_number = params.iter_number
        self._warmup_iter_number = params.warmup_iter_number
        self._ddpg_hparams_override = params.ddpg_hparams_dict
        self._hw_cfg_type = params.hw_cfg_type

    def apply_init(self) -> SingleConfigQuantizerSetup:
        from nncf.common.utils.debug import DEBUG_LOG_DIR
        from nncf.torch.automl.environment.quantization_env import QuantizationEnv

        if self._dump_autoq_data or is_debug():
            dump_dir = self._init_args.config.get("log_dir", None)
            if dump_dir is None:
                dump_dir = DEBUG_LOG_DIR
            self.dump_dir = Path(dump_dir) / Path("autoq") / Path("autoq_agent_dump")
            self.dump_dir.mkdir(parents=True, exist_ok=True)

            self.policy_dict = OrderedDict()  # key: episode
            self.best_policy_dict = OrderedDict()  # key: episode

            self._init_args.config["episodic_nncfcfg"] = str(self.dump_dir / "episodic_nncfcfg")
            os.makedirs(self._init_args.config["episodic_nncfcfg"], exist_ok=True)

            try:
                from torch.utils.tensorboard import SummaryWriter

                self.tb_writer = SummaryWriter(self.dump_dir)
                # log compression config to tensorboard
                self.tb_writer.add_text(
                    "AutoQ/run_config",
                    json.dumps(self._init_args.config["compression"], indent=4, sort_keys=False).replace("\n", "\n\n"),
                    0,
                )
            except ModuleNotFoundError:
                nncf_logger.warning(
                    "Tensorboard installation not found! Install tensorboard Python package "
                    "in order for AutoQ tensorboard statistics data to be dumped"
                )

        start_ts = datetime.now()

        from nncf.torch.automl.environment.quantization_env import QuantizationEnvParams

        env_params = QuantizationEnvParams(
            compression_ratio=self._params.compression_ratio,
            eval_subset_ratio=self._params.eval_subset_ratio,
            skip_constraint=self._params.skip_constraint,
            performant_bw=True,
            finetune=self._params.finetune,
            bits=self._params.bits,
            dump_init_precision_data=self._dump_autoq_data,
            log_dir=Path(DEBUG_LOG_DIR) / Path("autoq"),
        )

        # Instantiate Quantization Environment
        env = QuantizationEnv(
            self._model,
            self.quantization_controller,
            self._hw_precision_constraints,
            self._init_args.data_loader,
            self._init_args.eval_fn,
            hw_config_type=self._hw_cfg_type,
            params=env_params,
        )

        nb_state = len(env.state_list)
        nb_action = 1

        # Control buffer length at run manager level
        if "warmup_iter_number" not in self._ddpg_hparams_override:
            self._ddpg_hparams_override["warmup_iter_number"] = self._warmup_iter_number
        self._ddpg_hparams_override["rmsize"] = self._warmup_iter_number * (len(env.master_df) + 1)

        # Instantiate Automation Agent
        agent = DDPG(nb_state, nb_action, self._iter_number, hparam_override=self._ddpg_hparams_override)

        if self._dump_autoq_data and self.tb_writer is not None:
            # Need to replace '|' in nodestr (QuantizerId/QuantizerPointId)
            # to '+' as it is a special character in markdown
            temp_df = deepcopy(env.master_df[env.state_list + ["n_op"]])
            temp_df["modified_nodestr"] = list(map(lambda x: x.replace("|", "+"), temp_df.index.tolist()))
            temp_df = temp_df.set_index("modified_nodestr").reset_index()
            self.tb_writer.add_text("AutoQ/state_embedding", temp_df.to_markdown())

        best_policy, best_reward = self._search(agent, env)

        end_ts = datetime.now()

        final_qid_vs_qconfig_map = env.select_config_for_actions(best_policy)

        final_quantizer_setup = self.quantization_controller.get_quantizer_setup_for_current_state()
        for qp_id, qconf in final_qid_vs_qconfig_map.items():
            final_quantizer_setup.quantization_points[qp_id].qconfig = qconf

        str_bw = [str(element) for element in self.get_bitwidth_per_scope(final_quantizer_setup)]
        nncf_logger.info("\n".join(['[AutoQ]\n"bitwidth_per_scope": [', ",\n".join(str_bw), "]"]))
        nncf_logger.info(f"[AutoQ] best_reward: {best_reward}")
        nncf_logger.info(f"[AutoQ] best_policy: {best_policy}")
        nncf_logger.info("[AutoQ] Search completed.")
        nncf_logger.info("[AutoQ] Elapsed time of AutoQ Precision Initialization (): {}".format(end_ts - start_ts))
        return final_quantizer_setup

    def _search(self, agent: DDPG, env: "QuantizationEnv") -> Tuple[pd.Series, float]:  # noqa: F821
        best_reward = -math.inf
        episode = 0
        episode_reward = 0.0
        observation = None
        transition_buffer = []  # Transition buffer

        while episode < self._iter_number:  # counting based on episode
            episode_start_ts = time.time()
            if observation is None:
                # reset if it is the start of episode
                env.reset()
                observation = deepcopy(env.get_normalized_obs(0))
                agent.reset(observation)

            if episode < agent.warmup_iter_number:
                action = agent.random_action()
            else:
                action = agent.select_action(observation, episode=episode)

            # env response with next_observation, reward, terminate_info
            observation2, reward, done, info = env.step(map_precision(action))
            observation2 = deepcopy(observation2)
            transition_buffer.append([reward, deepcopy(observation), deepcopy(observation2), action, done])

            # update
            episode_reward += reward
            observation = deepcopy(observation2)

            if done:  # end of episode
                nncf_logger.info(
                    f"## Episode[{episode}], "
                    f"reward: {episode_reward:.3f}, "
                    f'acc: {info["accuracy"]:.3f}, '
                    f'model_ratio: {info["model_ratio"]:.3f}, '
                    f'model_size(MB): {info["model_size"] / 8e6:.2f}, '
                    f'BOP_ratio: {info["bop_ratio"]:.3f}\n'
                )

                # Replay Buffer Management
                if agent.memory.nb_entries % (len(env.master_df) + 1) > 0:
                    raise ValueError("logical bug in buffer management, uneven episode length")
                if agent.memory.limit % (len(env.master_df) + 1) > 0:
                    raise ValueError("replay buffer size must be divisible by episode step length")

                if agent.memory.nb_entries + len(transition_buffer) >= agent.memory.limit:
                    step_reward_per_episode = agent.memory.rewards.data[:: (len(transition_buffer) + 1)]
                    sorted_index_of_episodes = np.argsort(step_reward_per_episode)  # ascending order

                    # Retain the top 30% of highest rewarded episodes,
                    # discard by sampling an episode uniformly from the lower 70% episodes
                    discard_candidates = sorted_index_of_episodes[: int(len(sorted_index_of_episodes) * 0.7)]
                    discard_episode = np.random.choice(discard_candidates)

                    discard_start_index = (discard_episode) * (len(transition_buffer) + 1)
                    discard_end_index = (discard_episode + 1) * (len(transition_buffer) + 1)

                    agent.memory.discard(slice(discard_start_index, discard_end_index))
                # = EO Replay Buffer Management

                final_reward = transition_buffer[-1][0]
                r_per_step = final_reward / len(env.master_df)

                for i, (_, s_t, _, a_t, done) in enumerate(transition_buffer):
                    # Revision of prev_action as it could be modified by constrainer -------
                    if i == 0:
                        prev_action = 0.0
                    else:
                        prev_action = env.master_df["action"][i - 1] / 8  # ducktape scaling
                    if prev_action != s_t["prev_action"]:
                        s_t["prev_action"] = prev_action
                    # EO ------------------------
                    agent.observe(r_per_step, s_t, a_t, done)

                agent.memory.append(observation, agent.select_action(observation, episode=episode), 0.0, False)

                # update DDPG networks, note that this loop must not be
                # in the loop above to avoid non-numeric value in replay buffer
                for i, (_, s_t, _, a_t, done) in enumerate(transition_buffer):
                    if episode >= agent.warmup_iter_number:
                        for _ in range(agent.n_update):
                            agent.update_policy()

                # reset
                observation = None
                episode_reward = 0.0
                transition_buffer = []

                value_loss = agent.get_value_loss()
                policy_loss = agent.get_policy_loss()
                delta = agent.get_delta()

                nncf_stats = env.qctrl.statistics()
                bit_stats_df = (
                    pd.DataFrame.from_dict(
                        [nncf_stats.quantization.num_wq_per_bitwidth, nncf_stats.quantization.num_aq_per_bitwidth]
                    )
                    .fillna(0)
                    .astype(int)
                    .rename(index={0: "WQ", 1: "AQ"})
                    .transpose()
                    .sort_index(ascending=False)
                )
                bit_stats_df.index.name = "bitwidth"
                bit_stats_df = bit_stats_df.reset_index()

                if final_reward > best_reward:
                    best_reward = final_reward
                    best_policy = deepcopy(env.master_df["action"])
                    info_tuple = (episode, best_reward, info["accuracy"], info["model_ratio"], info["bop_ratio"])
                    self._dump_best_episode(info_tuple, bit_stats_df, env)
                    log_str = (
                        f"## Episode[{episode}] "
                        f"New best policy: {best_policy.values.tolist()}, "
                        f"reward: {best_reward:.3f}, "
                        f'acc: {info["accuracy"]:.3f}, '
                        f'model_ratio: {info["model_ratio"]:.3f},'
                        f' BOP_ratio: {info["bop_ratio"]:.3f}'
                    )
                    nncf_logger.info(f"\033[92m {log_str}\033[00m")

                episodic_info_tuple = (
                    episode,
                    final_reward,
                    best_reward,
                    info["accuracy"],
                    info["model_ratio"],
                    info["bop_ratio"],
                    value_loss,
                    policy_loss,
                    delta,
                )
                self._dump_episode(episodic_info_tuple, bit_stats_df, env, agent)

                episode_elapsed = time.time() - episode_start_ts

                nncf_logger.info(f'## Episode[{episode}] Policy: \n{env.master_df["action"].to_string()}\n')
                nncf_logger.info(f"## Episode[{episode}] Elapsed: {episode_elapsed:.3f}\n")

                episode += 1

        return best_policy, best_reward

    def _dump_best_episode(self, info_tuple: Tuple, bit_stats_df: pd.DataFrame, env: "QuantizationEnv"):  # noqa: F821
        if self._dump_autoq_data:
            episode = info_tuple[0]
            self.best_policy_dict[episode] = env.master_df["action"].astype("int")

            pd.DataFrame(self.best_policy_dict.values(), index=self.best_policy_dict.keys()).T.sort_index(
                axis=1, ascending=False
            ).to_csv(osp.join(self.dump_dir, "best_policy.csv"), index_label="nodestr")

            best_policy_string = self._generate_tensorboard_logging_string(
                bit_stats_df, env.master_df, info_tuple, env.skip_constraint
            )

            list_of_dump_dict = []
            for i, _ in enumerate(env.groups_of_adjacent_quantizers):
                list_of_dump_dict.append(env.master_df.loc[env.adjq_groupwise_df_lut_keys[i], ["action"]].to_dict())
            best_policy_string += (
                "\t\n\t# Precision(s) per Group of Adjacent Quantizers\n\t"
                + json.dumps(list_of_dump_dict, indent=4).replace("\n", "\n\t")
                + "\n\n"
            )

            self.tb_writer.add_text("AutoQ/best_policy", best_policy_string, episode)

    def _dump_episode(
        self, episodic_info_tuple: Tuple, bit_stats_df: pd.DataFrame, env: "QuantizationEnv", agent: DDPG  # noqa: F821
    ):
        if self._dump_autoq_data:
            episode, final_reward, _, accuracy, model_ratio, bop_ratio, _, _, _ = episodic_info_tuple

            current_bitwidth_per_scope = self.get_bitwidth_per_scope(env.qctrl.get_quantizer_setup_for_current_state())

            current_episode_nncfcfg = deepcopy(self._init_args.config)
            current_episode_nncfcfg["compression"]["initializer"]["precision"] = {
                "bitwidth_per_scope": current_bitwidth_per_scope
            }

            # Save nncf compression cfg
            episode_cfgfile = "{0}/{1:03d}_nncfcfg.json".format(
                str(self._init_args.config["episodic_nncfcfg"]), episode
            )

            with safe_open(Path(episode_cfgfile), "w") as outfile:
                json.dump(current_episode_nncfcfg, outfile, indent=4, sort_keys=False)

            self.policy_dict[episode] = env.master_df["action"].astype("int")
            pd.DataFrame(self.policy_dict.values(), index=self.policy_dict.keys()).T.sort_index(
                axis=1, ascending=False
            ).to_csv(osp.join(self.dump_dir, "policy_per_episode.csv"), index_label="nodestr")

            # log current episode policy and feedback as text
            info_tuple = (episode, final_reward, accuracy, model_ratio, bop_ratio)
            current_strategy_string = self._generate_tensorboard_logging_string(
                bit_stats_df, env.master_df, info_tuple, env.skip_constraint
            )

            if env.performant_bw is True:
                list_of_dump_dict = []
                for i, _ in enumerate(env.groups_of_adjacent_quantizers):
                    list_of_dump_dict.append(
                        env.master_df.loc[env.adjq_groupwise_df_lut_keys[i], ["action", "action_aligned"]].to_dict()
                    )
                current_strategy_string += (
                    "\t\n\t# Precision(s) per Group of Adjacent Quantizers\n\t"
                    + json.dumps(list_of_dump_dict, indent=4).replace("\n", "\n\t")
                    + "\n\n"
                )

            self.tb_writer.add_text("AutoQ/current_policy", current_strategy_string, episode)

            # visualization over episode
            if self.tb_writer is not None:
                self._add_to_tensorboard(self.tb_writer, episodic_info_tuple)

            if episode % int((self._iter_number + 10) / 10) == 0:
                agent.save_model(self.dump_dir)

    def _add_to_tensorboard(self, tb_writer: "SummaryWriter", log_tuple: Tuple):  # noqa: F821
        episode, final_reward, best_reward, accuracy, model_ratio, bop_ratio, value_loss, policy_loss, delta = log_tuple

        tb_writer.add_scalar("AutoQ/reward/last", final_reward, episode)
        tb_writer.add_scalar("AutoQ/reward/best", best_reward, episode)
        tb_writer.add_scalar("AutoQ/accuracy", accuracy, episode)
        tb_writer.add_scalar("AutoQ/model_ratio", model_ratio, episode)
        tb_writer.add_scalar("AutoQ/bop_ratio", bop_ratio, episode)
        tb_writer.add_scalar("AutoQ/agent/value_loss", value_loss, episode)
        tb_writer.add_scalar("AutoQ/agent/policy_loss", policy_loss, episode)
        tb_writer.add_scalar("AutoQ/agent/delta", delta, episode)

    def _generate_tensorboard_logging_string(
        self, bit_stats_df: pd.DataFrame, master_df: pd.DataFrame, info_tuple: Tuple, skip_constraint=False
    ) -> str:
        qdf = master_df  # For readibility
        episode, reward, accuracy, model_ratio, bop_ratio = info_tuple

        text_string = bit_stats_df.to_markdown() + "\n\n\n"
        text_string += "Episode: {:>4}, Reward: {:.3f}, ".format(episode, reward)
        text_string += "Accuracy: {:.3f}, Model_Size_Ratio: {:.3f}, BOP_Ratio: {:.3f}\n\n\n".format(
            accuracy, model_ratio, bop_ratio
        )

        for _, row_id in enumerate(qdf.index.tolist()):
            Qtype = "(WQ)" if qdf.is_wt_quantizer[row_id] else "(AQ)"

            if skip_constraint is False and qdf.loc[row_id, "action"] != qdf.loc[row_id, "unconstrained_action"]:
                text_string += "\t{} <= {} | {} {} \n".format(
                    str(int(qdf.loc[row_id, "action"])),
                    str(int(qdf.loc[row_id, "unconstrained_action"])),
                    Qtype,
                    row_id,
                )

            else:
                text_string += "\t{} | {} {} \n".format(str(int(qdf.loc[row_id, "action"])), Qtype, row_id)

        return text_string


def map_precision(action: float) -> int:
    precision_set = [2, 4, 8]
    precision_set = np.array(sorted(precision_set))
    tuned_point = precision_set + 3
    max_bit = max(precision_set)

    i = None
    for i, point in enumerate(tuned_point):
        if action <= 2**point / 2**max_bit:
            return int(precision_set[i])
    return int(precision_set[i])



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from dataclasses import fields
from typing import Any, ClassVar, Dict, List, Tuple

import nncf
from nncf.tensor import Tensor
from nncf.tensor import functions as fns


@dataclass
class TensorStatistic:
    """Base class that stores statistic data"""

    TENSOR_STATISTIC_OUTPUT_KEY: ClassVar[str] = "tensor_statistic_output"

    def get_data(self, is_serialized: bool = False) -> Dict[str, Any]:
        """
        Retrieves the data of the tensor statistics. If `is_serialized` is True,
        the data is prepared for serialization by including only Tensor instances.

        :param is_serialized: If True, the data is prepared for serialization by
            including only Tensor instances.
        :return: Dictionary with keys and their associated data. If `is_serialized`
            is True, the dictionary will contain only Tensor instances.
        """
        if is_serialized:
            return self._get_serialized_data()  # Dict[str, Tensor]
        return {field.name: getattr(self, field.name) for field in fields(self)}

    def _get_serialized_data(self) -> Dict[str, Tensor]:
        """
        Prepares the data for serialization by including only Tensor instances.

        :return: Dictionary with data for serialization.
        """
        serialized_data = {}
        for field in fields(self):
            key = field.name
            value = getattr(self, key)
            if isinstance(value, Tensor):
                serialized_data[key] = value
            else:
                raise nncf.InternalError(f"Unsupported type of value: {type(value)}")
        return serialized_data

    def load_data(self, data: Dict[str, Tensor]) -> None:
        """
        Loads the data from the serialized data.

        :param data: Data to load.
        """
        for key in (field.name for field in fields(self)):
            setattr(self, key, data[key])

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> TensorStatistic:
        args = {key: config[key] for key in (field.name for field in fields(cls))}
        return cls(**args)


@dataclass
class MinMaxTensorStatistic(TensorStatistic):
    MIN_STAT: ClassVar[str] = "min_values"
    MAX_STAT: ClassVar[str] = "max_values"

    min_values: Tensor
    max_values: Tensor

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, MinMaxTensorStatistic):
            return fns.allclose(self.min_values, other.min_values) and fns.allclose(self.max_values, other.max_values)
        return False


@dataclass
class AbsMaxTensorStatistic(TensorStatistic):
    ABS_MAX_STAT: ClassVar[str] = "abs_max"

    abs_max: Tensor

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, AbsMaxTensorStatistic):
            return fns.allclose(self.abs_max, other.abs_max)
        return False


@dataclass
class MeanTensorStatistic(TensorStatistic):
    MEAN_STAT: ClassVar[str] = "mean_values"
    SHAPE_STAT: ClassVar[str] = "shape"

    mean_values: Tensor
    shape: Tuple[int, ...]

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, MeanTensorStatistic):
            return self.shape == other.shape and fns.allclose(self.mean_values, other.mean_values)
        return False

    def _get_serialized_data(self) -> Dict[str, Tensor]:
        backend = self.mean_values.backend
        dtype = self.mean_values.dtype
        device = self.mean_values.device
        return {
            self.MEAN_STAT: self.mean_values,
            self.SHAPE_STAT: fns.tensor(self.shape, backend=backend, dtype=dtype, device=device),
        }

    def load_data(self, loaded_data: Dict[str, Tensor]) -> None:
        self.mean_values = loaded_data[self.MEAN_STAT]
        self.shape_values = tuple(loaded_data[self.SHAPE_STAT].tolist())


@dataclass
class MedianMADTensorStatistic(TensorStatistic):
    MEDIAN_VALUES_STAT: ClassVar[str] = "median_values"
    MAD_VALUES_STAT: ClassVar[str] = "mad_values"

    median_values: Tensor
    mad_values: Tensor

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, MedianMADTensorStatistic):
            return fns.allclose(self.median_values, other.median_values) and fns.allclose(
                self.mad_values, other.mad_values
            )
        return False

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> TensorStatistic:
        return cls(
            median_values=config[cls.TENSOR_STATISTIC_OUTPUT_KEY][cls.MEDIAN_VALUES_STAT],
            mad_values=config[cls.TENSOR_STATISTIC_OUTPUT_KEY][cls.MAD_VALUES_STAT],
        )


@dataclass
class PercentileTensorStatistic(TensorStatistic):
    PERCENTILE_VS_VALUE_DICT: ClassVar[str] = "percentile_vs_values_dict"

    percentile_vs_values_dict: Dict[str, Tensor]

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, PercentileTensorStatistic):
            if Counter(self.percentile_vs_values_dict.keys()) != Counter(other.percentile_vs_values_dict.keys()):
                return False
            for pct in self.percentile_vs_values_dict:
                if not fns.allclose(self.percentile_vs_values_dict[pct], other.percentile_vs_values_dict[pct]):
                    return False
            return True
        return False

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> TensorStatistic:
        if cls.TENSOR_STATISTIC_OUTPUT_KEY in config:
            percentile_vs_values_dict = config[cls.TENSOR_STATISTIC_OUTPUT_KEY]
        else:
            percentile_vs_values_dict = {}
            for (_, percentile), value in config.items():
                percentile_vs_values_dict[percentile] = value
        return cls(percentile_vs_values_dict=percentile_vs_values_dict)

    def _get_serialized_data(self) -> Dict[str, Tensor]:
        return self.PERCENTILE_VS_VALUE_DICT

    def load_data(self, loaded_data: Dict[str, Tensor]) -> None:
        self.percentile_vs_values_dict = loaded_data


@dataclass
class RawTensorStatistic(TensorStatistic):
    VALUES_STATS: ClassVar[str] = "values"

    values: Tensor

    def __eq__(self, other: RawTensorStatistic) -> bool:
        if isinstance(other, RawTensorStatistic):
            return fns.allclose(self.values, other.values)
        return False


@dataclass
class HessianTensorStatistic(TensorStatistic):
    HESSIAN_INPUT_ACTIVATION_STATS: ClassVar[str] = "hessian"

    hessian: Tensor

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, HessianTensorStatistic):
            return fns.allclose(self.hessian, other.hessian)
        return False


@dataclass
class MeanVarianceTensorStatistic(TensorStatistic):
    MEAN_VARIANCE_STAT: ClassVar[str] = "mean_variance"

    mean_variance: Tensor

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, MeanVarianceTensorStatistic):
            return fns.allclose(self.mean_variance, other.mean_variance)
        return False


@dataclass
class MaxVarianceTensorStatistic(TensorStatistic):
    MAX_VARIANCE_STAT: ClassVar[str] = "max_variance"

    max_variance: Tensor

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, MaxVarianceTensorStatistic):
            return fns.allclose(self.max_variance, other.max_variance)
        return False


@dataclass
class MeanMagnitudeTensorStatistic(TensorStatistic):
    MEAN_MAGNITUDE_STAT: ClassVar[str] = "mean_magnitude"

    mean_magnitude: Tensor

    def __eq__(self, other: TensorStatistic):
        if isinstance(other, MeanMagnitudeTensorStatistic):
            return fns.allclose(self.mean_magnitude, other.mean_magnitude)
        return False


@dataclass
class WCTensorStatistic(TensorStatistic):
    MEAN_STAT = "mean_values"
    SHAPE_STAT = "shape_values"

    mean_values: List[Tensor]
    shape_values: List[Tuple[Tensor]]

    def __eq__(self, other: Any) -> bool:
        shapes_equal = all(self.shapes[i] == other.shapes[i] for i in range(len(self.mean_values)))
        if not shapes_equal:
            return False
        mean_values_equal = all(
            self.tensor_eq(self.mean_values[i], other.mean_values[i]) for i in range(len(self.mean_values))
        )
        return mean_values_equal

    def _get_serialized_data(self) -> Dict[str, Tensor]:
        backend = self.mean_values[0].backend
        dtype = self.mean_values[0].dtype
        device = self.mean_values[0].device
        return {
            self.MEAN_STAT: fns.stack(self.mean_values),
            self.SHAPE_STAT: fns.tensor(
                [[dim.data for dim in shape] for shape in self.shape_values],
                backend=backend,
                dtype=dtype,
                device=device,
            ),
        }

    def load_data(self, loaded_data: Dict[str, Tensor]) -> None:
        self.shape_values = [tuple(shape) for shape in loaded_data[self.SHAPE_STAT]]
        self.mean_values = [it for it in loaded_data[self.MEAN_STAT]]

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> TensorStatistic:
        mean_values, shape_values = None, None
        if cls.MEAN_STAT in config and config[cls.MEAN_STAT] is not None:
            mean_values = [fns.squeeze(it) for it in config[cls.MEAN_STAT]]
        if cls.SHAPE_STAT in config and config[cls.SHAPE_STAT] is not None:
            shape_values = [tuple(it) for it in config[cls.SHAPE_STAT]]
        return cls(mean_values=mean_values, shape_values=shape_values)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar

import torch

from nncf.common.logging import nncf_logger
from nncf.torch.dynamic_graph.context import TracingContext
from nncf.torch.dynamic_graph.graph import DynamicGraph
from nncf.torch.dynamic_graph.io_handling import FillerInputInfo
from nncf.torch.dynamic_graph.io_handling import LoaderInputInfo
from nncf.torch.dynamic_graph.io_handling import ModelInputInfo
from nncf.torch.dynamic_graph.wrappers import wrap_parameters
from nncf.torch.utils import get_model_device
from nncf.torch.utils import is_multidevice


class GraphTracer:
    def __init__(self, custom_forward_fn: Callable[[torch.nn.Module], Any]):
        self.custom_forward_fn = custom_forward_fn

    def trace_graph(
        self,
        model: torch.nn.Module,
        context_to_use: Optional[TracingContext] = None,
        as_eval: bool = False,
        trace_parameters: bool = False,
    ) -> DynamicGraph:
        if context_to_use is None:
            context_to_use = TracingContext()

        context_to_use.enable_trace_dynamic_graph()
        from nncf.torch.utils import training_mode_switcher

        with context_to_use as _ctx:
            _ctx.base_module_thread_local_replica = model
            with torch.no_grad():
                if trace_parameters:
                    wrap_parameters(model)

                if as_eval:
                    with training_mode_switcher(model, is_training=False):
                        self.custom_forward_fn(model)
                else:
                    self.custom_forward_fn(model)

        context_to_use.disable_trace_dynamic_graph()
        return context_to_use.graph


T = TypeVar("T")
WrapInputsFnType = Callable[[Tuple, Dict], Tuple[Tuple, Dict]]
WrapOutputsFnType = Callable[[T], T]


def create_dummy_forward_fn(
    input_info: ModelInputInfo,
    with_input_tracing: bool = False,
    wrap_inputs_fn: WrapInputsFnType = None,
    wrap_outputs_fn: WrapOutputsFnType = None,
    with_output_tracing: bool = False,
):
    def default_dummy_forward_fn(model):
        from nncf.torch.dynamic_graph.io_handling import replicate_same_tensors
        from nncf.torch.dynamic_graph.io_handling import wrap_nncf_model_inputs_with_objwalk
        from nncf.torch.dynamic_graph.io_handling import wrap_nncf_model_outputs_with_objwalk

        device = None
        if isinstance(input_info, (FillerInputInfo, LoaderInputInfo)):
            if is_multidevice(model):
                nncf_logger.warning(
                    "Multidevice model detected when tracing the model's dynamic graph - will pass example "
                    "inputs to the model as-is without changing their device."
                )
            else:
                device = get_model_device(model)
        args, kwargs = input_info.get_forward_inputs(device)

        if with_input_tracing:
            if wrap_inputs_fn is None:
                # We control the input argument structure w.r.t. tensors if input_info is a FillerInputInfo
                # - a simple objwalk application should be sufficient in this simple case.
                # For more control, wrap_inputs_fn is used when this is used in NNCFNetwork
                # which is guaranteed to be the same as during the actual NNCFNetwork.forward
                args, kwargs = wrap_nncf_model_inputs_with_objwalk(args, kwargs)
            else:
                args, kwargs = wrap_inputs_fn(args, kwargs)
        retval = model(*args, **kwargs)
        if with_output_tracing:
            retval = replicate_same_tensors(retval)
            if wrap_outputs_fn is not None:
                return wrap_outputs_fn(retval)
            return wrap_nncf_model_outputs_with_objwalk(retval)
        return retval

    return default_dummy_forward_fn



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Callable, Dict, TypeVar

TObj = TypeVar("TObj", bound=type)


class StatefulClassesRegistry:
    """
    Registry for the stateful classes  - classes that can be restored from their state by `from_state` method.
    """

    REQUIRED_METHOD_NAME = "from_state"

    def __init__(self) -> None:
        self._name_vs_class_map: Dict[str, type] = {}
        self._class_vs_name_map: Dict[type, str] = {}

    def register(self, name: str = None) -> Callable[[TObj], TObj]:
        """
        Decorator to map class with some name - specified in the argument or name of the class.

        :param name: The registration name. By default, it's name of the class.
        :return: The inner function for registration.
        """

        def decorator(cls: TObj) -> TObj:
            registered_name = name if name is not None else cls.__name__

            if registered_name in self._name_vs_class_map:
                raise ValueError(
                    "{} has already been registered to {}".format(
                        registered_name, self._name_vs_class_map[registered_name]
                    )
                )

            if cls in self._class_vs_name_map:
                raise ValueError("{} has already been registered to {}".format(cls, self._class_vs_name_map[cls]))

            if inspect.isclass(cls) and not hasattr(cls, self.REQUIRED_METHOD_NAME):
                raise ValueError(
                    "Cannot register a class ({}) that does not have {}() method.".format(
                        registered_name, self.REQUIRED_METHOD_NAME
                    )
                )

            self._class_vs_name_map[cls] = registered_name
            self._name_vs_class_map[registered_name] = cls

            return cls

        return decorator

    def get_registered_class(self, registered_name: str) -> type:
        """
        Provides a class that was registered with the given name.

        :param registered_name: name
        :return: class that was registered with the given name
        """
        if registered_name in self._name_vs_class_map:
            return self._name_vs_class_map[registered_name]
        raise KeyError("No registered stateful classes with {} name".format(registered_name))

    def get_registered_name(self, stateful_cls: type) -> str:
        """
        Provides a name that was used to register the given stateful class.

        :param stateful_cls: class
        :return: name that was used on registration of the given class
        """
        if stateful_cls in self._class_vs_name_map:
            return self._class_vs_name_map[stateful_cls]
        raise KeyError("The class {} was not registered.".format(stateful_cls.__name__))


class CommonStatefulClassesRegistry:
    """
    Common for TF and PT registry for the stateful classes.
    """

    @staticmethod
    def register(name: str = None) -> Callable[[TObj], TObj]:
        """
        Decorator to map class with some name - specified in the argument or name of the class.

        :param name: The registration name. By default, it's name of the class.
        :return: The inner function for registration.
        """

        def decorator(cls: TObj) -> TObj:
            PT_STATEFUL_CLASSES.register(name)(cls)
            TF_STATEFUL_CLASSES.register(name)(cls)
            return cls

        return decorator

    @staticmethod
    def get_registered_class(registered_name: str) -> type:
        """
        Provides a class that was registered with the given name.

        :param registered_name: name
        :return: class that was registered with the given name
        """
        return PT_STATEFUL_CLASSES.get_registered_class(registered_name)

    @staticmethod
    def get_registered_name(stateful_cls: type) -> str:
        """
        Provides a name that was used to register the given stateful class.

        :param stateful_cls: class
        :return: name that was used on registration of the given class
        """
        return PT_STATEFUL_CLASSES.get_registered_name(stateful_cls)


PT_STATEFUL_CLASSES = StatefulClassesRegistry()
TF_STATEFUL_CLASSES = StatefulClassesRegistry()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import tensorflow as tf

from nncf.common.logging import nncf_logger
from nncf.common.statistics import NNCFStatistics


class StatisticsCallback(tf.keras.callbacks.Callback):
    """
    Callback for logging compression statistics to tensorboard and stdout.
    """

    def __init__(
        self,
        statistics_fn: Callable[[], NNCFStatistics],
        log_tensorboard: bool = True,
        log_text: bool = True,
        log_dir: str = None,
    ):
        """
        Initializes compression statistics callback.

        :param statistics_fn: A callable object that provides NNCF statistics.
        :param log_tensorboard: Whether to log statistics to tensorboard or not.
        :param log_text: Whether to log statistics to stdout.
        :param log_dir: The directory for tensorbard logging.
        """
        super().__init__()
        self._statistics_fn = statistics_fn
        self._log_tensorboard = log_tensorboard
        self._log_text = log_text
        self._file_writer = None
        if log_tensorboard:
            if log_dir is None:
                raise ValueError("log_dir must be specified if log_tensorboard is true.")

            self._file_writer = tf.summary.create_file_writer(log_dir + "/compression")

    def _dump_to_tensorboard(self, logs: dict, step: int):
        with self._file_writer.as_default():
            for name, value in logs.items():
                tf.summary.scalar(name, value, step=step)

    def on_epoch_end(self, epoch: int, logs: dict = None):
        nncf_stats = self._statistics_fn()
        if self._log_tensorboard:
            self._dump_to_tensorboard(
                self._prepare_for_tensorboard(nncf_stats), self.model.optimizer.iterations.numpy()
            )
        if self._log_text:
            nncf_logger.info(nncf_stats.to_str())

    def on_train_begin(self, logs: dict = None):
        nncf_stats = self._statistics_fn()
        if self._log_tensorboard:
            self._dump_to_tensorboard(
                self._prepare_for_tensorboard(nncf_stats), self.model.optimizer.iterations.numpy()
            )
        if self._log_text:
            nncf_logger.info(nncf_stats.to_str())

    def on_train_end(self, logs: dict = None):
        if self._file_writer:
            self._file_writer.close()

    def _prepare_for_tensorboard(self, stats: NNCFStatistics):
        raise NotImplementedError(
            "StatisticsCallback class implementation must override the _prepare_for_tensorboard method."
        )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Optional, Tuple, Type

import numpy as np
import openvino.runtime as ov
import openvino.runtime.op as op
import openvino.runtime.opset13 as opset

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import GenericWeightedLayerAttributes
from nncf.common.graph.layer_attributes import LinearLayerAttributes
from nncf.common.graph.layer_attributes import WeightedLayerAttributes
from nncf.common.tensor_statistics.collectors import ReductionAxes
from nncf.openvino.graph.layout import OVLayoutElem
from nncf.openvino.graph.layout import get_conv_weights_layout
from nncf.openvino.graph.layout import get_conv_weights_layout_from_node
from nncf.openvino.graph.layout import get_linear_activations_layout_from_node
from nncf.openvino.graph.layout import get_linear_input_layout
from nncf.openvino.graph.layout import get_linear_weights_layout_from_node
from nncf.openvino.graph.metatypes.groups import CONV_OPERATIONS
from nncf.openvino.graph.metatypes.groups import OPERATIONS_WITH_BIAS
from nncf.openvino.graph.metatypes.groups import OPERATIONS_WITH_WEIGHTS
from nncf.openvino.graph.metatypes.openvino_metatypes import OVAddMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConstantMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConvertMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConvolutionBackpropDataMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVGroupConvolutionBackpropDataMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVIfMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVMatMulMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVOpMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import get_node_metatype
from nncf.tensor import Tensor
from nncf.tensor import TensorBackend

InplaceInsertionFnType = Callable[[ov.Node, int, str], ov.Node]


def get_add_bias_node(node: NNCFNode, nncf_graph: NNCFGraph) -> Optional[NNCFNode]:
    """
    Returns Add node which stores bias for node.

    :param node: NNCFGraph node.
    :param nncf_graph: NNCFGraph.
    :return: Add node if exists.
    """
    for child in nncf_graph.get_next_nodes(node):
        if child.metatype == OVAddMetatype:
            bias_constant = get_node_with_bias_value(child, nncf_graph)
            if bias_constant is not None:
                return child
    return None


def is_node_with_bias(
    node: NNCFNode, nncf_graph: NNCFGraph, metatypes_with_bias: Optional[List[OVOpMetatype]] = None
) -> bool:
    """
    Checks if the node has a bias or not.

    :param node: The node to check.
    :param nncf_graph: NNCFGraph instance.
    :param metatypes_with_bias: List of the metatypes that contains biases.
    :return: Return `True` if `node` corresponds to the operation
        with bias (bias is added to the output tensor of that operation),
        `False` otherwise.
    """
    if metatypes_with_bias is None:
        metatypes_with_bias = OPERATIONS_WITH_BIAS

    if node.metatype not in metatypes_with_bias:
        return False

    # Since we do not verify bias constant shape, we need to verify the weight's existence at least.
    # layer_attributes adds only for nodes with weights, according to the nncf_graph_builder.py for the backend.
    if node.layer_attributes is None:
        return False

    return get_add_bias_node(node, nncf_graph) is not None


def get_number_if_op(model: ov.Model) -> int:
    """
    Returns number of If operation in a model.

    :param model: Model.
    :return: True if Model has If operation, False - otherwise.
    """

    def cnt_if_op(model: ov.Model, cnt: int) -> int:
        for model_op in model.get_ops():
            if get_node_metatype(model_op) == OVIfMetatype:
                cnt += 1
                cnt = cnt_if_op(model_op.get_function(0), cnt)
                cnt = cnt_if_op(model_op.get_function(1), cnt)
        return cnt

    return cnt_if_op(model, 0)


def get_const_value(const_node: ov.Node, cast_bf16_to_fp32: bool = True) -> np.ndarray:
    """
    Returns the constant tensor for the node.
    This method is applicable only for the floating-point constant data.

    :param const_node: OpenVINO node.
    :param cast_bf16_to_fp32: Whether to cast bf16 node data to fp32 or not. If False and the node contains bf16 data,
        the resulting bf16 value will be returned encoded inside a numpy.float16 array.
    :return: The constant value.
    """
    if const_node.get_element_type() == ov.Type.bf16 and cast_bf16_to_fp32:
        return const_node.get_data(dtype=np.float32)
    return const_node.data


def get_bias_value(
    node_with_bias: NNCFNode, nncf_graph: NNCFGraph, model: ov.Model, node_mapping: Dict[str, ov.Node] = None
) -> np.ndarray:
    """
    Returns the bias tensor for the biased node.

    :param node_with_bias: The node that corresponds to the operation with bias.
    :param nncf_graph: NNCFGraph instance.
    :param model: The model that contains this operation.
    :param node_mapping: Original nodes mapping cache.
    :return: The bias value that is applied to the output tensor of the node's operation.
    """
    if node_mapping is None:
        node_mapping = {op.get_friendly_name(): op for op in model.get_ops()}
    bias_constant = get_node_with_bias_value(get_add_bias_node(node_with_bias, nncf_graph), nncf_graph)
    ov_bias_constant = node_mapping[bias_constant.node_name]
    return get_const_value(ov_bias_constant)


def get_weight_value(node_with_weight: NNCFNode, model: ov.Model, port_id: int) -> np.ndarray:
    """
    Returns a weight value for the node with weight.

    :param node_with_weight: Node with weight.
    :param nncf_graph: NNCF graph.
    :param model: The model that contains this operation.
    :param port_id: The input port ID to get weight input.
    :return: The weight value.
    """
    const_op_friendly_name = node_with_weight.layer_attributes.constant_attributes[port_id]["name"]
    friendly_name_to_op_map = {op.get_friendly_name(): op for op in model.get_ops()}
    const_op = friendly_name_to_op_map[const_op_friendly_name]
    weight_tensor = get_const_value(const_op)
    return weight_tensor


def get_node_with_bias_value(add_node: NNCFNode, nncf_graph: NNCFGraph) -> Optional[NNCFNode]:
    """
    Returns node that represents bias constant in the NNCF graph, if it exists.

    :param add_node: NNCFNode that provides bias.
    :param nncf_graph: NNCFGraph instance.
    :return: Optional NNCFNode with bias value.
    """
    if add_node.layer_attributes is None:
        return None

    const_port_ids = add_node.layer_attributes.get_const_port_ids()
    assert len(const_port_ids) == 1
    bias_port_id = const_port_ids[0]
    bias_constant = nncf_graph.get_input_edge_by_port_id(add_node, bias_port_id).from_node

    if bias_constant.metatype == OVConvertMetatype:
        bias_constant = nncf_graph.get_input_edge_by_port_id(bias_constant, 0).from_node

    return bias_constant if bias_constant.metatype == OVConstantMetatype else None


def get_result_node_name(output_name: str, port_id: int) -> str:
    """
    Returns name of Result based on node name and its port.

    :param output_name: Node name.
    :param port_id: Node port.
    :return: Name of Result.
    """

    return f"Result_{output_name}.{port_id}"


def get_parameter_node_name(parameter_name: str, port_id: int) -> str:
    """
    Returns name of Parameter based on node name and its port.

    :param parameter_name: Node name.
    :param port_id: Node port.
    :return: Name of Parameter.
    """

    return f"Parameter_{parameter_name}.{port_id}"


def get_ov_model_reduce_node_name(output_name: str, reduce_node_name: str, port_id: int) -> str:
    """
    Returns name of reduce node based on output name, node type and port id.

    :param output_name: Target node name.
    :param node_type: Reduce node name.
    :param port_id: Target port id of the target node.
    :return: Reduce node name.
    """
    return f"{output_name}_{reduce_node_name}.{port_id}"


def get_inplace_reduce_op(
    op: Type[ov.Node],
    reduction_axes: Optional[ReductionAxes],
    use_abs: bool,
    keep_dims: bool = True,
) -> InplaceInsertionFnType:
    """
    Returns inplace insertion function that adds reduce node to a passed node.

    :param op: OpenVINO reduction operation type to insert.
    :param reduction_axes: Target reduction axes for the reduction node.
        Reduce along all axes in case reduction_axes are None.
    :param use_abs: Whether reduce absolute values of input tensors or not.
    :param keep_dims: Whether to keep the original dimension length or return result as a scalar.
    :returns: Inplace insertion function to use in ModelTransformer.
    """

    def get_reduce_op(node: ov.Node, output_port_id: int, output_node_name: str) -> ov.Node:
        reduction_axes_ = reduction_axes
        if reduction_axes_ is None:
            partial_shape = get_partial_shape_safe(node, output_port_id)
            reduction_axes_ = np.arange(partial_shape.rank.get_length()).astype(np.int64)

        if use_abs:
            op_input = opset.abs(node.output(output_port_id), name="abs_" + output_node_name)
            output_port_id = 0
        else:
            op_input = node

        return op(
            op_input.output(output_port_id),
            reduction_axes=np.array(reduction_axes_, dtype=np.int64),
            keep_dims=keep_dims,
            name=output_node_name,
        )

    return get_reduce_op


def get_inplace_min_op(reduction_axes: Optional[ReductionAxes]) -> InplaceInsertionFnType:
    """
    Returns inplace min function that adds reduce min node to a passed node.

    :param reduction_axes: Target reduction axes for the reduction node.
        Reduce along all axes in case reduction_axes are None.
    :returns: Inplace insertion function to use in ModelTransformer.
    """
    return get_inplace_reduce_op(opset.reduce_min, reduction_axes, False)


def get_inplace_max_op(
    reduction_axes: Optional[ReductionAxes], use_abs_max: bool, keep_dims: bool = True
) -> InplaceInsertionFnType:
    """
    Returns inplace max function that adds reduce max node to a passed node.

    :param reduction_axes: Target reduction axes for the reduction node.
        Reduce along all axes in case reduction_axes are None.
    :param use_abs_max: Whether reduce absolute values of input tensors or not.
    :param keep_dims: Whether to keep the original dimension length or return result as a scalar.
    :returns: Inplace insertion function to use in ModelTransformer.
    """
    return get_inplace_reduce_op(opset.reduce_max, reduction_axes, use_abs_max, keep_dims)


def get_inplace_mean_op(reduction_axes: Optional[ReductionAxes]) -> InplaceInsertionFnType:
    """
    Returns inplace mean function that adds reduce mean node to a passed node.

    :param reduction_axes: Target reduction axes for the reduction node.
        Reduce along all axes in case reduction_axes are None.
    :returns: Inplace insertion function to use in ModelTransformer.
    """
    return get_inplace_reduce_op(opset.reduce_mean, reduction_axes, False)


def var_op(
    op_input: ov.Output, output_node_name: str, reduction_axes: Optional[np.ndarray] = None, keep_dims: bool = True
) -> ov.Node:
    """
    Return a subgraph computing variance on a given output.

    :param op_input: An output to compute variance for.
    :param output_node_name: Variance output name.
    :param keep_dims: Whether to keep the original dimension length or return result as a scalar.
    :param reduction_axes: Axes along which to compute variance.
    """
    mean = opset.reduce_mean(
        op_input,
        reduction_axes=reduction_axes,
        keep_dims=True,
        name=f"{output_node_name}/mean",
    )
    diff = opset.squared_difference(mean, op_input, name=f"{output_node_name}/squared_diff")
    variance = opset.reduce_mean(
        diff,
        reduction_axes=reduction_axes,
        keep_dims=keep_dims,
        name=output_node_name,
    )
    return variance


def get_inplace_mean_var_op(reduction_axes: Optional[ReductionAxes] = None) -> InplaceInsertionFnType:
    """
    Return an operation getter function that computes variance across given axes and then mean-reduces the result across
    the remaining axes.

    :param reduction_axes: Axes along which to compute variance.
    """

    def get_mean_var_reduce_op(node: ov.Node, output_port_id: int, output_node_name: str) -> ov.Node:
        partial_shape = get_partial_shape_safe(node, output_port_id)
        all_axes = np.arange(partial_shape.rank.get_length()).astype(np.int64)
        reduction_axes_ = np.array(all_axes if reduction_axes is None else reduction_axes, dtype=np.int64)

        reduce_all = np.array_equal(reduction_axes_, all_axes)
        var_op_name = output_node_name if reduce_all else f"{output_node_name}/var"
        result = var_op(node.output(output_port_id), var_op_name, reduction_axes_, keep_dims=not reduce_all)
        if not reduce_all:
            result = opset.reduce_mean(
                result,
                reduction_axes=all_axes,
                keep_dims=False,
                name=output_node_name,
            )

        return result

    return get_mean_var_reduce_op


def get_inplace_max_var_op(reduction_axes: Optional[ReductionAxes] = None) -> InplaceInsertionFnType:
    """
    Return an operation getter function that computes variance across given axes and then max-reduces the result across
    the remaining axes.

    :param reduction_axes: Axes along which to compute variance.
    """

    def get_max_var_reduce_op(node: ov.Node, output_port_id: int, output_node_name: str) -> ov.Node:
        partial_shape = get_partial_shape_safe(node, output_port_id)
        all_axes = np.arange(partial_shape.rank.get_length()).astype(np.int64)
        reduction_axes_ = np.array(all_axes if reduction_axes is None else reduction_axes, dtype=np.int64)

        reduce_all = np.array_equal(reduction_axes_, all_axes)
        var_op_name = output_node_name if reduce_all else f"{output_node_name}/var"
        result = var_op(node.output(output_port_id), var_op_name, reduction_axes_, keep_dims=not reduce_all)
        if not reduce_all:
            result = opset.reduce_max(
                result,
                reduction_axes=all_axes,
                keep_dims=False,
                name=output_node_name,
            )

        return result

    return get_max_var_reduce_op


def get_inplace_mean_max_op(reduction_axes: Optional[ReductionAxes], use_abs_max: bool) -> InplaceInsertionFnType:
    """
    Return an operation getter function that computes maximum across given axes and then mean-reduces the result across
    the remaining axes.

    :param reduction_axes: Axes to compute maximum across.
    :param use_abs_max: Whether to apply abs() operation before the max operation.
    """

    def get_mean_max_reduce_op(node: ov.Node, output_port_id: int, output_node_name: str) -> ov.Node:
        partial_shape = get_partial_shape_safe(node, output_port_id)
        all_axes = np.arange(partial_shape.rank.get_length()).astype(np.int64)
        reduction_axes_ = np.array(all_axes if reduction_axes is None else reduction_axes, dtype=np.int64)

        reduce_all = np.array_equal(reduction_axes_, all_axes)
        max_op_name = output_node_name if reduce_all else f"{output_node_name}/max"
        result = get_inplace_max_op(reduction_axes, use_abs_max, keep_dims=not reduce_all)(
            node, output_port_id, max_op_name
        )
        if not reduce_all:
            result = opset.reduce_mean(
                result,
                reduction_axes=all_axes,
                keep_dims=False,
                name=output_node_name,
            )

        return result

    return get_mean_max_reduce_op


def get_inplace_shape_op() -> InplaceInsertionFnType:
    """
    Return an operation returning a shape on the given output.
    """

    def get_shape_op(node: ov.Node, output_port_id: int, output_node_name: str) -> ov.Node:
        result = opset.shape_of(node.output(output_port_id), name=output_node_name)
        return result

    return get_shape_op


def get_inplace_batch_mean_op() -> InplaceInsertionFnType:
    """
    Returns inplace batch mean function that adds reduce batch mean node to a passed node.

    :returns: Inplace insertion function to use in ModelTransformer.
    """
    return get_inplace_reduce_op(opset.reduce_mean, np.array(0), False)


def get_inplace_mean_per_ch(axis: int) -> InplaceInsertionFnType:
    """
    Returns inplace mean per channel function that adds reduce mean per channel node
    to a passed node.

    :param axis: Channel axis.
    :returns: Inplace insertion function to use in ModelTransformer.
    """

    def get_reduce_op(node: ov.Node, output_port_id: int, output_node_name: str) -> ov.Node:
        input_shape = get_partial_shape_safe(node, output_port_id)
        input_shape = [dim.get_length() if dim.is_static else -1 for dim in input_shape]
        if len(input_shape) < 3:
            return opset.reduce_mean(
                node.output(output_port_id),
                reduction_axes=0,
                keep_dims=False,
                name=output_node_name,
            )

        ch_dim = 1
        if axis != ch_dim:
            transpose_dims = list(range(len(input_shape)))
            transpose_dims[axis], transpose_dims[ch_dim] = transpose_dims[ch_dim], transpose_dims[axis]
            transposed_shape = [input_shape[dim] for dim in transpose_dims]

            reshape_input_node = opset.transpose(node.output(output_port_id), transpose_dims)
            output_port_id = 0
        else:
            reshape_input_node = node
            transposed_shape = input_shape

        keeped_dims = transposed_shape[:2]
        keeped_dims = [0 if dim < 0 else dim for dim in keeped_dims]
        squized_dims = -1 if -1 in transposed_shape[2:] else np.prod(transposed_shape[2:])
        reshape_op = opset.reshape(
            reshape_input_node.output(output_port_id),
            output_shape=np.array((keeped_dims[0], keeped_dims[1], squized_dims)),
            special_zero=True,
        )
        return opset.reduce_mean(
            reshape_op,
            reduction_axes=np.array((0, 2)),
            keep_dims=False,
            name=output_node_name,
        )

    return get_reduce_op


def get_partial_shape_safe(node, port_id) -> Tuple[int, ...]:
    partial_shape = node.get_output_partial_shape(port_id)
    if partial_shape.rank.is_dynamic or not partial_shape.all_non_negative:
        raise nncf.ValidationError(
            f"Could not collect statistics for the node {node} because its output shape rank is dynamic or negative"
        )
    return partial_shape


def get_reducer_output_node_names(
    node_type, target_node_name: str, port_id: int, fn_output_port_id: int, inplace: bool
) -> List[str]:
    """
    Returns output name to feed to a reducer node.

    :param node_type: String that describes reduce node type.
    :param target_node_name: Name of the node inputs/outputs/weights of which was
        used for reduction.
    :param port_id: Target port id of the target node.
    :param fn_output_port_id: Port id of the reducer subgraph.
    :param inplace: Wheather reducer calculated inplace or not.
    :return: Output names to feed to a reducer node.
    """
    if inplace:
        target_node_name = get_ov_model_reduce_node_name(target_node_name, node_type, port_id)
        return [get_result_node_name(target_node_name, fn_output_port_id)]
    return [get_result_node_name(target_node_name, port_id)]


def get_weight_channel_axes(node: NNCFNode) -> List[int]:
    """
    Returns axes numbers of the weight tensor which correspond to its channels.

    :param node: NNCFNode with weights.
    :param weights_port_id: Weight port id of the target node.
    :return: Axes numbers of the weight tensor which correspond to its channels.
    """
    if node.metatype not in OPERATIONS_WITH_WEIGHTS:
        raise ValueError("Channel axis cannot be defined for operation without weights.")

    if node.metatype in CONV_OPERATIONS:
        weights_layout = get_conv_weights_layout_from_node(node)
        return [idx for idx, elem in enumerate(weights_layout) if elem in [OVLayoutElem.GROUPS, OVLayoutElem.C_OUT]]
    elif node.metatype == OVMatMulMetatype:
        return get_matmul_channel_axes(node)
    return node.metatype.const_channel_axis


def get_matmul_channel_axes(node: ov.Node) -> List[int]:
    """
    Calculate channel axes for the MatMul operation.

    :param node: The target node.
    :return: List of channel axes for the MatMul operation.
    """
    weights_layout = get_linear_weights_layout_from_node(node)
    return [idx for idx, elem in enumerate(weights_layout) if elem in [OVLayoutElem.SPATIAL, OVLayoutElem.C_OUT]]


def create_bias_tensor(node_without_bias: NNCFNode, graph: NNCFGraph, value: Any) -> np.ndarray:
    """
    Creates bias value constant array filled by given value.

    :param node_without_bias: NNCFNode to add bias to.
    :param graph: Target NNCFgraph.
    :param value: Value to fill bias constant array.
    :return: Bias value constant array filled by given value.
    """
    node_shape = graph.get_output_edges(node_without_bias)[0].tensor_shape
    bias_shape = [1] * len(node_shape)
    channel_axis = node_without_bias.metatype.output_channel_axis
    bias_shape[channel_axis] = node_shape[1]
    return np.full(bias_shape, value)


def get_weighted_layer_attributes(
    ov_node: ov.Node, ov_metatype: OVOpMetatype, constant_attributes: Dict[int, Any]
) -> WeightedLayerAttributes:
    """
    Function retrieves common layer attributes from the given node.

    :param ov_node: TargetOpenvino graph node instance.
    :param ov_metatype: NNCF Openvino metatype of the given node.
    :param constant_attributes: Constant attributes collected for the given node.
    :return: Weighted layer attributes for the given node.
    """
    if len(constant_attributes) != 1:
        return None

    port_id, attrs = constant_attributes.copy().popitem()
    if ov_metatype in CONV_OPERATIONS:
        node_attrs = ov_node.get_attributes()
        kwargs = {
            "weight_requires_grad": False,
            "stride": tuple(node_attrs["strides"]),
            "dilations": node_attrs["dilations"],
            "transpose": ov_metatype in [OVConvolutionBackpropDataMetatype, OVGroupConvolutionBackpropDataMetatype],
            # TODO: ticket 114378: unify pad attribute
            "padding_values": tuple(node_attrs["pads_begin"] + node_attrs["pads_end"]),
        }
        weights_shape = attrs["shape"]
        weights_layout = get_conv_weights_layout(ov_metatype=ov_metatype, weights_shape=weights_shape)
        kwargs.update(
            {
                "in_channels": weights_shape[weights_layout.index(OVLayoutElem.C_IN)],
                "out_channels": weights_shape[weights_layout.index(OVLayoutElem.C_OUT)],
                "kernel_size": tuple(
                    dim for dim, elem in zip(weights_shape, weights_layout) if elem == OVLayoutElem.SPATIAL
                ),
                "groups": (
                    weights_shape[weights_layout.index(OVLayoutElem.GROUPS)]
                    if OVLayoutElem.GROUPS in weights_layout
                    else 1
                ),
            }
        )

        return ConvolutionLayerAttributes(**kwargs)
    if ov_metatype == OVMatMulMetatype:
        weights_shape = attrs["shape"]
        weights_layout = get_linear_input_layout(
            input_shape=weights_shape, transpose=attrs["transpose"], port_id=port_id
        )

        kwargs = {
            "weight_requires_grad": False,
            "in_features": weights_shape[weights_layout.index(OVLayoutElem.C_IN)],
            "out_features": (
                weights_shape[weights_layout.index(OVLayoutElem.C_OUT)]
                if OVLayoutElem.C_OUT in weights_layout
                else None
            ),
            "with_bias": False,
        }
        return LinearLayerAttributes(**kwargs)
    return GenericWeightedLayerAttributes(weight_requires_grad=False, weight_shape=attrs.get("shape", None))


def get_activation_channel_axis(node: NNCFNode, port_id: int, input_shape: Tuple[int]) -> int:
    """
    Returns axis number of the activation tensor which correspond to it channel.

    :param node: NNCFNode instance.
    :param port_id: Port ID for input.
    :param input_shape: Shape of the input.
    :return: Channel axis number.
    """
    # In case of the OpenVINO, [N, C, ..] layout applicable for most quantizable layers.
    channel_axis = 1

    # But the MatMul layers may transpose inputs internally.
    if node.metatype == OVMatMulMetatype:
        activations_layout = get_linear_activations_layout_from_node(node, port_id, input_shape)
        channel_axis = activations_layout.index(OVLayoutElem.C_IN)

    return channel_axis


def convert_op(node: ov.Node, target_dtype: ov.Type) -> ov.Node:
    """
    Return a subgraph which converts the given node output to the target data type. If the output is already in the
    target data type then the given node is returned.

    :param node: The input node to convert.
    :param target_dtype: The target data type to convert the input node to.
    :return: The converted node.
    """
    if node.get_element_type() == target_dtype:
        return node
    return opset.convert(node, target_dtype)


def non_convertable_divide_op(a: ov.Node, b: ov.Node) -> ov.Node:
    """
    Creates a "non-convertable" divide operation. It won't be converted to a*(1/b).
    """
    divide_node = a / b
    divide_node.get_rt_info()["nonconvertable_divide_0"] = True
    return divide_node


def create_ov_const_from_tensor(x: Tensor, dtype: ov.Type, name: Optional[str] = None) -> op.Constant:
    """
    Create an OpenVINO Constant node from the given tensor.
    :param x: Data tensor. Supports NumPy and OV tensor backends. If x backend is OV, the constant node is created
        directly from underlying OV tensor.
    :param dtype: Data type of the constant.
    :param name: Optional name of the constant.
    :return: OpenVINO Constant node.
    """
    if x.backend == TensorBackend.ov:
        assert x.data.get_element_type() == dtype
        return opset.constant(x.data, name=name, shared_memory=True)
    const = opset.constant(x.data, dtype=dtype, name=name)
    return const



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.definitions import ONLINE_DOCS_ROOT
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.defaults import NUM_BN_ADAPTATION_SAMPLES

BATCHNORM_ADAPTATION_SCHEMA = {
    "type": "object",
    "description": f"This initializer is applied by default to utilize batch norm statistics adaptation to the "
    f"current compression scenario. See "
    f"[documentation]"
    f"({ONLINE_DOCS_ROOT}docs/compression_algorithms/Quantization.md#batch-norm-statistics-adaptation) "
    f"for more details.",
    "properties": {
        "num_bn_adaptation_samples": with_attributes(
            NUMBER,
            description="Number of samples from the training "
            "dataset to use for model inference "
            "during the BatchNorm statistics "
            "adaptation procedure for the compressed "
            "model. The actual number of samples will "
            "be a closest multiple of the batch "
            "size. Set this to 0 to disable BN adaptation.",
            default=NUM_BN_ADAPTATION_SAMPLES,
        )
    },
    "additionalProperties": False,
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import path as osp
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
from torch.distributed import barrier
from torch.nn import Module

import nncf
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.compression import BaseCompressionAlgorithmController as BaseController
from nncf.common.deprecation import warning_deprecated
from nncf.common.logging import nncf_logger
from nncf.common.utils.api_marker import api
from nncf.common.utils.debug import set_debug_log_dir
from nncf.config import NNCFConfig
from nncf.config.extractors import extract_algorithm_names
from nncf.config.extractors import has_input_info_field
from nncf.config.telemetry_extractors import CompressionStartedFromConfig
from nncf.telemetry import tracked_function
from nncf.telemetry.events import NNCF_PT_CATEGORY
from nncf.telemetry.extractors import FunctionCallTelemetryExtractor
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.algo_selector import NoCompressionAlgorithmBuilder
from nncf.torch.composite_compression import PTCompositeCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.dynamic_graph.graph_tracer import WrapInputsFnType
from nncf.torch.dynamic_graph.graph_tracer import WrapOutputsFnType
from nncf.torch.dynamic_graph.io_handling import EXTRA_STRUCTS_WITH_DATALOADERS
from nncf.torch.dynamic_graph.io_handling import ExampleInputInfo
from nncf.torch.dynamic_graph.io_handling import FillerInputInfo
from nncf.torch.dynamic_graph.io_handling import LoaderInputInfo
from nncf.torch.dynamic_graph.io_handling import ModelInputInfo
from nncf.torch.graph.transformations.serialization import deserialize_transformations
from nncf.torch.model_transformer import PTModelTransformer
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.utils import is_dist_avail_and_initialized
from nncf.torch.utils import is_main_process
from nncf.torch.utils import maybe_convert_legacy_names_in_compress_state
from nncf.torch.utils import training_mode_switcher


@api(canonical_alias="nncf.torch.create_compressed_model")
@tracked_function(
    NNCF_PT_CATEGORY,
    [
        CompressionStartedFromConfig(argname="config"),
    ],
)
def create_compressed_model(
    model: Module,
    config: NNCFConfig,
    compression_state: Optional[Dict[str, Any]] = None,
    dummy_forward_fn: Callable[[Module], Any] = None,
    wrap_inputs_fn: Callable[[Tuple, Dict], Tuple[Tuple, Dict]] = None,
    wrap_outputs_fn: Callable[[Any], Any] = None,
    dump_graphs=True,
) -> Tuple[CompressionAlgorithmController, NNCFNetwork]:
    """
    The main function used to produce a model ready for compression fine-tuning from an original PyTorch
    model and a configuration object.

    :param model: The original model. Should have its parameters already loaded from a checkpoint or another
        source.
    :param config: A configuration object used to determine the exact compression modifications to be applied
        to the model
    :type config: nncf.NNCFConfig
    :param compression_state: representation of the entire compression state to unambiguously restore
        the compressed model. Includes builder and controller states.
    :param dummy_forward_fn: if supplied, will be used instead of a *forward* function call to build
        the internal graph representation via tracing. Specifying this is useful when the original training pipeline
        has special formats of data loader output or has additional *forward* arguments other than input tensors.
        Otherwise, the *forward* call of the model during graph tracing will be made with mock tensors according
        to the shape specified in the config object. The dummy_forward_fn code MUST contain calls to
        nncf.nncf_model_input functions made with each compressed model input tensor in the underlying model's
        args/kwargs tuple, and these calls should be exactly the same as in the wrap_inputs_fn function code
        (see below); if dummy_forward_fn is specified, then wrap_inputs_fn also must be specified.
    :param wrap_inputs_fn: if supplied, will be used on the module's input arguments during a regular, non-dummy
        forward call before passing the inputs to the underlying compressed model. This is required if the model's
        input tensors that are important for compression are not supplied as arguments to the model's forward call
        directly, but instead are located in a container (such as list), and the model receives the container as an
        argument. wrap_inputs_fn should take as input two arguments - the tuple of positional arguments to the
        underlying model's forward call, and a dict of keyword arguments to the same. The function should wrap each
        tensor among nncf.nncf_model_input function, which is a no-operation function and marks the tensors as inputs
        to be traced by NNCF in the internal graph representation. Output is the tuple of (args, kwargs), where args
        and kwargs are the same as were supplied in input, but each tensor in the original input. Must be specified
        if dummy_forward_fn is specified.
    :param wrap_outputs_fn: same as `wrap_inputs_fn`, but applies to model outputs
    :param dump_graphs: Whether to dump the internal graph representation of the
        original and compressed models in the .dot format into the log directory.
    :return: A controller for the compression algorithm (or algorithms, in which case the controller
        is an instance of CompositeCompressionController) and the model ready for compression parameter training wrapped
        as an object of NNCFNetwork.
    """

    warning_deprecated(
        "The 'nncf.torch.create_compressed_model' function is deprecated and will be removed in a future release.\n"
        "To perform post training quantization (PTQ) or quantization aware training (QAT),"
        " use the nncf.quantize() API:\n"
        " - https://github.com/openvinotoolkit/nncf?tab=readme-ov-file#post-training-quantization\n"
        " - https://github.com/openvinotoolkit/nncf?tab=readme-ov-file#training-time-quantization\n"
        "Examples:\n"
        " - https://github.com/openvinotoolkit/nncf/tree/develop/examples/post_training_quantization/torch\n"
        " - https://github.com/openvinotoolkit/nncf/tree/develop/examples/quantization_aware_training/torch"
    )

    if isinstance(model, NNCFNetwork):
        raise nncf.InternalError(
            "The model object has already been compressed.\n"
            "NNCF for PyTorch modifies the model object in-place, and repeat calls to "
            "`nncf.torch.create_compressed_model` with the same model object passed as argument "
            "will lead to an incorrect attempt to compress the model twice.\n"
            "Make sure that the model object you are passing has not already been compressed (for "
            "instance, by testing `if isinstance(model, nncf.torch.nncf_network.NNCFNetwork)`).\n"
            "If you are encountering this in a Jupyter notebook context - make sure that when "
            "re-running cells involving `nncf.torch.create_compressed_model` the original model object "
            "is also re-created (via constructor call)."
        )

    set_debug_log_dir(config.get("log_dir", "."))

    is_legacy_model_state_dict = (
        compression_state is not None
        and BaseController.BUILDER_STATE not in compression_state
        and BaseController.CONTROLLER_STATE not in compression_state
    )
    maybe_convert_legacy_names_in_compress_state(compression_state)

    should_init = compression_state is None

    nncf_network = create_nncf_network(model, config, dummy_forward_fn, wrap_inputs_fn, wrap_outputs_fn)

    if dump_graphs and is_main_process():
        nncf_network.nncf.get_graph().visualize_graph(osp.join(config.get("log_dir", "."), "original_graph.dot"))
    builder = create_compression_algorithm_builder(config, should_init)

    is_state_loadable = not is_legacy_model_state_dict and compression_state is not None
    if is_state_loadable:
        builder.load_state(compression_state[BaseController.BUILDER_STATE])
    compressed_model = builder.apply_to(nncf_network)
    compression_ctrl = builder.build_controller(compressed_model)

    if is_state_loadable:
        compression_ctrl.load_state(compression_state[BaseController.CONTROLLER_STATE])

    compressed_model.nncf.set_compression_controller(compression_ctrl)

    # Required to ensure that the model leaving create_compressed_model has correct compressed graph.
    # In particular, this is currently required for correct functioning of RNNs.
    compressed_model.nncf.rebuild_graph()

    try:
        if is_legacy_model_state_dict:
            from nncf.torch import load_state

            state_dict_to_load = compression_state.get("state_dict", compression_state)
            load_state(compressed_model, state_dict_to_load, is_resume=True)
    finally:
        if dump_graphs and is_main_process():
            compressed_model_graph = compressed_model.nncf.get_graph()
            compressed_model_graph.visualize_graph(osp.join(config.get("log_dir", "."), "compressed_graph.dot"))

    synchronize_all_processes_in_distributed_mode()
    return compression_ctrl, compressed_model


def get_input_info_from_config(config: NNCFConfig) -> ModelInputInfo:
    if has_input_info_field(config):
        return FillerInputInfo.from_nncf_config(config)

    nncf_logger.debug(
        "Config has no 'input_info' section, trying to use dataloader output as model inputs " "for graph building."
    )
    exact_info = LoaderInputInfo.from_nncf_config_dataloaders(config)
    if exact_info is not None:
        return exact_info
    raise nncf.ValidationError(
        "Could not determine tensor inputs for the model's forward call.\n"
        "If you are using the `nncf.quantize` API, make sure that you supply the "
        "calibration dataloader to the `nncf.quantize` call.\n"
        "If you are using the `create_compressed_model` API, either specify the "
        "inputs by using the 'input_info' section in the NNCFConfig, or attach an "
        "initialization dataloader to the NNCFConfig by calling "
        "`NNCFConfig.register_extra_structs(...)` with one of the following extra "
        f"structures:\n"
        f"{EXTRA_STRUCTS_WITH_DATALOADERS}\n"
        f"or by calling `nncf.torch.register_default_init_args`"
    )


def create_nncf_network(
    model: torch.nn.Module,
    config: NNCFConfig,
    dummy_forward_fn: Callable[[Module], Any] = None,
    wrap_inputs_fn: WrapInputsFnType = None,
    wrap_outputs_fn: WrapOutputsFnType = None,
) -> NNCFNetwork:
    """
    The main function used to produce a model ready for adding compression from an original PyTorch
    model and a configuration object.

    :param model: The original model. Should have its parameters already loaded from a checkpoint or another
        source.
    :param config: A configuration object used to determine the exact compression modifications to be applied
        to the model
    :param dummy_forward_fn: if supplied, will be used instead of a *forward* function call to build
        the internal graph representation via tracing. Specifying this is useful when the original training pipeline
        has special formats of data loader output or has additional *forward* arguments other than input tensors.
        Otherwise, the *forward* call of the model during graph tracing will be made with mock tensors according
        to the shape specified in the config object. The dummy_forward_fn code MUST contain calls to
        nncf.nncf_model_input
        functions made with each compressed model input tensor in the underlying model's args/kwargs tuple, and these
        calls should be exactly the same as in the wrap_inputs_fn function code (see below); if dummy_forward_fn is
        specified, then wrap_inputs_fn also must be specified.
    :param wrap_inputs_fn: if supplied, will be used on the module's input arguments during a regular, non-dummy
        forward call before passing the inputs to the underlying compressed model. This is required if the model's input
        tensors that are important for compression are not supplied as arguments to the model's forward call directly,
        but instead are located in a container (such as list), and the model receives the container as an argument.
        wrap_inputs_fn should take as input two arguments - the tuple of positional arguments to the underlying
        model's forward call, and a dict of keyword arguments to the same. The function should wrap each tensor among
        the supplied model's args and kwargs that is important for compression (e.g. quantization) with an
        nncf.nncf_model_input function, which is a no-operation function and marks the tensors as inputs to be traced
        by NNCF in the internal graph representation. Output is the tuple of (args, kwargs), where args and kwargs are
        the same as were supplied in input, but each tensor in the original input. Must be specified if
        dummy_forward_fn is specified.
    :param wrap_outputs_fn: Same as `wrap_inputs_fn`, but for marking model outputs with

    :return: A model wrapped by NNCFNetwork, which is ready for adding compression."""

    if dummy_forward_fn is not None and wrap_inputs_fn is None:
        raise ValueError(
            "A custom dummy forward function was specified, but the corresponding input wrapping function "
            "was not. In case a custom dummy forward function is specified for purposes of NNCF graph "
            "building, then the wrap_inputs_fn parameter MUST also be specified and be consistent with "
            "the input wrapping done in dummy_forward_fn."
        )

    # Preserve `.training`/`.requires_grad` state since we will be building NNCFNetwork in `.eval` mode
    with training_mode_switcher(model, is_training=False):
        # Compress model that will be deployed for the inference on target device. No need to compress parts of the
        # model that are used on training stage only (e.g. AuxLogits of Inception-v3 model) or unused modules with
        # weights. As a consequence, no need to care about spoiling BN statistics, as they're disabled in eval mode.

        input_info = get_input_info_from_config(config)
        scopes_without_shape_matching = config.get("scopes_without_shape_matching", [])
        ignored_scopes = config.get("ignored_scopes")
        target_scopes = config.get("target_scopes")

        nncf_network = NNCFNetwork(
            model,
            input_info=input_info,
            dummy_forward_fn=dummy_forward_fn,
            wrap_inputs_fn=wrap_inputs_fn,
            wrap_outputs_fn=wrap_outputs_fn,
            ignored_scopes=ignored_scopes,
            target_scopes=target_scopes,
            scopes_without_shape_matching=scopes_without_shape_matching,
        )

        nncf_network.nncf.get_tracing_context().disable_trace_dynamic_graph()

    synchronize_all_processes_in_distributed_mode()
    return nncf_network


def synchronize_all_processes_in_distributed_mode():
    if is_dist_avail_and_initialized():
        try:
            barrier()
        # Exception can be raised during running barrier
        # if the backend not in the supported list https://pytorch.org/docs/stable/distributed.html
        except RuntimeError as err:
            nncf_logger.warning(
                "Training pipeline spawned an error while synchronizing distributed training processes:"
            )
            nncf_logger.warning(err)
            nncf_logger.warning("Desynchronization of distributed processes may occur.")


def create_compression_algorithm_builder(config: NNCFConfig, should_init=True) -> PTCompressionAlgorithmBuilder:
    """
    Create compression algorithm builders by a given list of algorithm names.

    :param config: A configuration object used to determine the exact compression modifications to be applied
    to the model
    :param should_init: The flag indicates that the generated compression builder will initialize (True) or not (False)
    the training parameters of the model during model building.
    :return: compression algorithm builder
    """
    algo_names = extract_algorithm_names(config)
    return create_compression_algorithm_builder_from_algo_names(algo_names, config, should_init)


def create_compression_algorithm_builder_from_algo_names(
    algo_names: List[str], config: NNCFConfig, should_init: bool
) -> PTCompressionAlgorithmBuilder:
    """
    Create compression algorithm builders by a given list of algorithm names.

    :param algo_names: list of algorithm names
    :param config: A configuration object used to determine the exact compression modifications to be applied
    to the model
    :param should_init: The flag indicates that the generated compression builder will initialize (True) or not (False)
    the training parameters of the model during model building.
    :return: compression algorithm builder
    """
    if not algo_names:
        algo_builder_classes = [NoCompressionAlgorithmBuilder]
    else:
        algo_builder_classes = [PT_COMPRESSION_ALGORITHMS.get(algo_name) for algo_name in algo_names]
    if len(algo_builder_classes) == 1:
        builder = next(iter(algo_builder_classes))(config, should_init=should_init)
    else:
        builder = PTCompositeCompressionAlgorithmBuilder(config, should_init=should_init)
    return builder


@tracked_function(
    NNCF_PT_CATEGORY,
    [
        FunctionCallTelemetryExtractor("nncf.torch.wrap_model"),
    ],
)
def wrap_model(
    model: torch.nn.Module,
    example_input: Any,
    trace_parameters: bool = False,
) -> NNCFNetwork:
    """
    Wraps a PyTorch model to the NNCFNetwork class.

    This function dynamically extends the instance of PyTorch model with NNCF-enabling functionality.

    :param model: PyTorch model.
    :param example_input: An example input that will be used for model tracing. A tuple is interpreted
        as an example input of a set of non keyword arguments, and a dict as an example input of a set
        of keywords arguments.
    :param trace_parameters: Whether to trace model parameters. Default is False.
    :return: A model wrapped by NNCFNetwork.
    """
    if not isinstance(model, torch.nn.Module):
        raise TypeError(
            f"The provided model type {type(model)} is incompatible. "
            "Only models inheriting from torch.nn.Module are supported."
        )

    input_info = ExampleInputInfo.from_example_input(example_input)

    with training_mode_switcher(model, is_training=False):
        nncf_network = NNCFNetwork(
            model, input_info=input_info, replace_modules=not trace_parameters, trace_parameters=trace_parameters
        )
        nncf_network.nncf.get_tracing_context().disable_trace_dynamic_graph()

    return nncf_network


def is_wrapped_model(model: torch.nn.Module) -> bool:
    """
    Check that the model was wrapped by NNCFNetwork.

    :param model: A model.
    :return: True if the model is wrapped, False otherwise.
    """
    return isinstance(model, NNCFNetwork)


@tracked_function(
    NNCF_PT_CATEGORY,
    [
        FunctionCallTelemetryExtractor("nncf.torch.load_from_config"),
    ],
)
def load_from_config(model: torch.nn.Module, config: Dict[str, Any], example_input: Any) -> NNCFNetwork:
    """
    Wraps given model to a NNCFNetwork and recovers additional modules from given NNCFNetwork config.
    Does not recover additional modules weights as they are located in a corresponded state_dict.

    :param model: PyTorch model.
    :param config: NNCNetwork config.
    :param example_input: An example input that will be used for model tracing. A tuple is interpreted
        as an example input of a set of non keyword arguments, and a dict as an example input of a set
        of keywords arguments.
    :return: NNCFNetwork builded from given model with additional modules recovered from given NNCFNetwork config.
    """
    nncf_network = wrap_model(model, example_input, trace_parameters=config.pop(NNCFNetwork.TRACE_PARAMETERS_KEY))
    transformation_layout = deserialize_transformations(config)
    return PTModelTransformer(nncf_network).transform(transformation_layout)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Any, Callable, Dict, List, Union

import torch

from nncf.common.graph import NNCFNodeName
from nncf.common.graph.transformations.commands import Command
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.transformations.commands import TransformationType

DEFAULT_HOOKS_GROUP_NAME = "default_hooks_group"


class PTTargetPointStateNames:
    TARGET_NODE_NAME = "target_node_name"
    INPUT_PORT = "input_port_id"
    TARGET_TYPE = "target_type"


class PTTargetPoint(TargetPoint):
    _OPERATION_TYPES = [
        TargetType.PRE_LAYER_OPERATION,
        TargetType.POST_LAYER_OPERATION,
        TargetType.OPERATION_WITH_WEIGHTS,
    ]
    _HOOK_TYPES = [TargetType.OPERATOR_PRE_HOOK, TargetType.OPERATOR_POST_HOOK]

    _LAYER_TYPE = [TargetType.LAYER]

    _state_names = PTTargetPointStateNames

    def __init__(self, target_type: TargetType, target_node_name: NNCFNodeName, *, input_port_id: int = None):
        super().__init__(target_type)
        self.target_node_name = target_node_name
        self.target_type = target_type
        if self.target_type not in self._OPERATION_TYPES + self._HOOK_TYPES + self._LAYER_TYPE:
            raise NotImplementedError("Unsupported target type: {}".format(target_type))

        self.input_port_id = input_port_id

    def __eq__(self, other: "PTTargetPoint"):
        return (
            isinstance(other, PTTargetPoint)
            and self.target_type == other.target_type
            and self.target_node_name == other.target_node_name
            and self.input_port_id == other.input_port_id
        )

    def __str__(self):
        prefix = str(self.target_type)
        retval = prefix
        if self.target_type in self._OPERATION_TYPES + self._LAYER_TYPE:
            retval += " {}".format(self.target_node_name)
        elif self.target_type in self._HOOK_TYPES:
            if self.input_port_id is not None:
                retval += " {}".format(self.input_port_id)
            retval += " " + str(self.target_node_name)
        return retval

    def __hash__(self):
        return hash(str(self))

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.TARGET_TYPE: self.target_type.get_state(),
            self._state_names.INPUT_PORT: self.input_port_id,
            self._state_names.TARGET_NODE_NAME: self.target_node_name,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "PTTargetPoint":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        kwargs = {
            cls._state_names.TARGET_TYPE: TargetType.from_state(state[cls._state_names.TARGET_TYPE]),
            cls._state_names.INPUT_PORT: state[cls._state_names.INPUT_PORT],
            cls._state_names.TARGET_NODE_NAME: state[cls._state_names.TARGET_NODE_NAME],
        }
        return cls(**kwargs)


class PTCommand(Command):
    """
    The base class for all Command for PyTorch.
    """

    def requires_graph_rebuild(self):
        """
        Return boolean flag to rebuild graph of model.

        :return: Boolean flag.
        """
        return False


class PTTransformationCommand(TransformationCommand):
    """
    The base class for all TransformationCommand for PyTorch.
    """

    def requires_graph_rebuild(self):
        """
        Return boolean flag to rebuild graph of model.

        :return: Boolean flag.
        """
        return False


class PTInsertionCommand(PTTransformationCommand):
    """
    Insertion operation to the models.
    """

    def __init__(
        self,
        point: PTTargetPoint,
        fn: Callable,
        priority: Union[TransformationPriority, int] = TransformationPriority.DEFAULT_PRIORITY,
        hooks_group_name: str = DEFAULT_HOOKS_GROUP_NAME,
    ):
        super().__init__(TransformationType.INSERT, point)
        self.fn: Callable = fn
        self.priority: TransformationPriority = priority
        self.hooks_group_name = hooks_group_name

    def requires_graph_rebuild(self):
        # Rebuild graph when adding quantization nodes.
        return self.priority == TransformationPriority.QUANTIZATION_PRIORITY


class ExtraCompressionModuleType(Enum):
    EXTERNAL_QUANTIZER = 0
    EXTERNAL_OP = 1


class PTSharedFnInsertionCommand(PTTransformationCommand):
    def __init__(
        self,
        target_points: List[PTTargetPoint],
        fn: Callable,
        op_unique_name: str,
        compression_module_type: ExtraCompressionModuleType = ExtraCompressionModuleType.EXTERNAL_OP,
        priority: Union[TransformationPriority, int] = TransformationPriority.DEFAULT_PRIORITY,
        hooks_group_name: str = DEFAULT_HOOKS_GROUP_NAME,
    ):
        super().__init__(TransformationType.INSERT, None)
        self.target_points = target_points
        self.fn = fn
        self.op_name = op_unique_name
        self.compression_module_type = compression_module_type
        self.priority = priority
        self.hooks_group_name = hooks_group_name

    def requires_graph_rebuild(self):
        return True


class PTModelExtractionCommand(PTCommand):
    """
    Extracts submodel based on the sub-model input and output names
    """

    def __init__(self, input_node_names: List[str], output_node_names: List[str]):
        """
        :param node_name: Node name that will be extracted.
        """
        super().__init__(TransformationType.EXTRACT)
        self.input_node_names = input_node_names
        self.output_node_names = output_node_names


class PTBiasCorrectionCommand(PTTransformationCommand):
    """
    Corrects bias value in the model based on the input value.
    """

    def __init__(self, target_point: PTTargetPoint, bias_value: torch.Tensor):
        """
        :param target_point: The TargetPoint instance for the correction that contains layer's information.
        :param bias_value: The bias shift value that will be added to the original bias value.
        """
        super().__init__(TransformationType.CHANGE, target_point)
        self.bias_value = bias_value


class PTWeightUpdateCommand(PTTransformationCommand):
    """
    Corrects weight value in the model based on the input value.
    """

    def __init__(self, target_point: PTTargetPoint, weight_value: torch.Tensor):
        """
        :param target_point: The TargetPoint instance for the correction that contains layer's information.
        :param weight_value: The new weight value that will be used instead of the original weight value.
        """
        super().__init__(TransformationType.CHANGE, target_point)
        self.weight_value = weight_value



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from nncf.api.statistics import Statistics
from nncf.common.utils.api_marker import api


@api()
@dataclass
class TrainingLoopStatistics(Statistics):
    """
    Contains statistics related to Accuracy Aware Training Loop
    """

    uncompressed_accuracy: float
    compression_rate: float
    compressed_accuracy: float
    absolute_accuracy_degradation: float
    relative_accuracy_degradation: float
    accuracy_budget: float

    def to_str(self) -> str:
        stats_str = (
            f"Uncompressed model accuracy: {self.uncompressed_accuracy:.4f}\n"
            f"Compressed model accuracy: {self.compressed_accuracy:.4f}\n"
            f"Model compression rate: {self.compression_rate:.4f}\n"
            f"Absolute accuracy drop: {self.absolute_accuracy_degradation:.4f}\n"
            f"Relative accuracy drop: {self.relative_accuracy_degradation:.2f}%\n"
            f"Accuracy budget: {self.accuracy_budget:.4f}"
        )
        return stats_str



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict, List, Optional

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.utils import get_first_nodes_of_type
from nncf.common.logging import nncf_logger
from nncf.common.pruning.clusterization import Cluster
from nncf.common.pruning.clusterization import Clusterization
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.model_analysis import ModelAnalyzer
from nncf.common.pruning.model_analysis import cluster_special_ops
from nncf.common.pruning.utils import PruningAnalysisDecision
from nncf.common.pruning.utils import PruningAnalysisReason
from nncf.common.pruning.utils import PruningOperationsMetatypeRegistry
from nncf.common.pruning.utils import get_output_channels
from nncf.common.pruning.utils import get_previous_convs
from nncf.common.pruning.utils import get_sources_of_node
from nncf.common.pruning.utils import is_batched_linear
from nncf.common.pruning.utils import is_conv_with_downsampling
from nncf.common.pruning.utils import is_grouped_conv
from nncf.common.pruning.utils import is_prunable_depthwise_conv
from nncf.common.scopes import should_consider_scope


class PruningNodeSelector:
    """
    Determines which of the nodes with pruning types should be pruned
    and which of them should be pruned together.
    """

    def __init__(
        self,
        pruning_operator_metatypes: PruningOperationsMetatypeRegistry,
        prune_operations_types: List[str],
        grouping_operations_types: List[str],
        ignored_scopes: Optional[List[str]],
        target_scopes: Optional[List[str]],
        prune_first: bool,
        prune_downsample_convs: bool,
    ):
        """
        :param pruning_operator_metatypes: registry with operation metatypes pruning algorithm is aware of, i.e.
            metatypes describing operations with common pruning mask application and propagation properties, e.g.
            IdentityMaskForwardOps unifies operations that propagate pruning masks as is (relu, swish etc.), whereas
            Convolution unifies different convolution operations (conv1d, conv2d, conv3d) which accepts some input masks
            and provide some output masks.
        :param prune_operations_types: Types of operations with prunable parameters.
        :param grouping_operations_types: Types of operations causing the need to prune connected to them
            operations together.
        :param ignored_scopes: Ignored scopes.
        :param target_scopes: Target scopes.
        :param prune_first: Whether to prune first convolution or not.
        :param prune_downsample_convs: Whether to prune downsample Convolutional layers (with stride > 1) or not.
        """
        self._pruning_operator_metatypes = pruning_operator_metatypes
        pruning_op_metatypes_dict = self._pruning_operator_metatypes.registry_dict
        self._identity_mask_propagation_op_metatype = pruning_op_metatypes_dict["identity_mask_propagation"]
        self._stop_propagation_op_metatype = pruning_op_metatypes_dict["stop_propagation_ops"]

        self._prune_operations_types = prune_operations_types
        self._grouping_operations_types = grouping_operations_types

        self._ignored_scopes = ignored_scopes
        self._target_scopes = target_scopes

        self._prune_first = prune_first
        self._prune_downsample_convs = prune_downsample_convs

    def create_pruning_groups(self, graph: NNCFGraph) -> Clusterization[NNCFNode]:
        """
        This function groups ALL nodes with pruning types to groups that should be pruned together.
            1. Create clusters for special ops (eltwises) that should be pruned together
            2. Create groups of nodes that should be pruned together (taking into account clusters of special ops)
            3. Add remaining single nodes
            4. Unite clusters for Conv + Depthwise conv (should be pruned together too)
            5. Checks for groups (all nodes in group can prune or all group can't be pruned)
        Return groups of modules that should be pruned together.

        :param graph: Graph to work with and their initialization parameters as values.
        :return: Clusterization of pruned nodes.
        """

        all_nodes_to_prune = graph.get_nodes_by_types(self._prune_operations_types)  # NNCFNodes here

        # 1. Clusters for special ops
        identity_like_types = self._identity_mask_propagation_op_metatype.get_all_op_aliases()
        special_ops_clusterization = cluster_special_ops(graph, self._grouping_operations_types, identity_like_types)

        pruned_nodes_clusterization = Clusterization[NNCFNode](lambda x: x.node_id)

        # 2. Clusters for nodes that should be pruned together (taking into account clusters for special ops)
        for i, cluster in enumerate(special_ops_clusterization.get_all_clusters()):
            all_pruned_inputs: Dict[int, NNCFNode] = {}
            clusters_to_merge: List[int] = []

            for node in cluster.elements:
                sources = get_sources_of_node(node, graph, self._prune_operations_types)
                for source_node in sources:
                    if pruned_nodes_clusterization.is_node_in_clusterization(source_node.node_id):
                        # Merge clusters if some node already added in another cluster
                        cluster = pruned_nodes_clusterization.get_cluster_containing_element(source_node.node_id)
                        clusters_to_merge.append(cluster.id)
                    elif source_node.node_id not in all_pruned_inputs:
                        all_pruned_inputs[source_node.node_id] = source_node

            if all_pruned_inputs:
                cluster = Cluster[NNCFNode](i, list(all_pruned_inputs.values()), list(all_pruned_inputs.keys()))
                clusters_to_merge.append(cluster.id)
                pruned_nodes_clusterization.add_cluster(cluster)

            # Merge clusters if one source node in several special ops clusters
            pruned_nodes_clusterization.merge_list_of_clusters(clusters_to_merge)

        last_cluster_idx = len(special_ops_clusterization.get_all_clusters())

        # 3. Add remaining single nodes as separate clusters
        for node in all_nodes_to_prune:
            if not pruned_nodes_clusterization.is_node_in_clusterization(node.node_id):
                cluster = Cluster[NNCFNode](last_cluster_idx, [node], [node.node_id])
                pruned_nodes_clusterization.add_cluster(cluster)

                last_cluster_idx += 1

        stop_propagation_ops = self._stop_propagation_op_metatype.get_all_op_aliases()
        # 4. Merge clusters for Conv + Depthwise conv (should be pruned together too)
        for node in all_nodes_to_prune:
            cluster_id = pruned_nodes_clusterization.get_cluster_containing_element(node.node_id).id

            if is_prunable_depthwise_conv(node):
                previous_convs = get_previous_convs(graph, node, self._prune_operations_types, stop_propagation_ops)
                previous_clusters = [
                    pruned_nodes_clusterization.get_cluster_containing_element(node.node_id).id
                    for node in previous_convs
                ]
                pruned_nodes_clusterization.merge_list_of_clusters([cluster_id] + previous_clusters)

        # 5. Merge nodes into one cluster if some module forwards several times
        multiforward_nodes = self._get_multiforward_nodes(graph)
        for list_of_nodes in multiforward_nodes:
            clusters_to_merge = [
                pruned_nodes_clusterization.get_cluster_containing_element(node.node_id).id for node in list_of_nodes
            ]
            pruned_nodes_clusterization.merge_list_of_clusters(clusters_to_merge)

            # Merge previous convolutions into one cluster
            all_previous_convs = []
            for node in list_of_nodes:
                nncf_node = graph.get_node_by_id(node.node_id)
                previous_convs = get_previous_convs(
                    graph, nncf_node, self._prune_operations_types, stop_propagation_ops
                )
                # Check if previous node isn't multiforward,
                # in case of multiforward nodes cycle
                for previous_conv in previous_convs:
                    if previous_conv not in list_of_nodes:
                        all_previous_convs.append(previous_conv)

            previous_clusters = [
                pruned_nodes_clusterization.get_cluster_containing_element(node.node_id).id
                for node in all_previous_convs
            ]
            pruned_nodes_clusterization.merge_list_of_clusters(previous_clusters)

        # 6. Checks for groups (all nodes in group can be pruned or all group can't be pruned).
        model_analyser = ModelAnalyzer(graph, self._pruning_operator_metatypes, self._prune_operations_types)
        can_prune_analysis = model_analyser.analyse_model_before_pruning()
        can_prune_and_should_prune_analysis = self._should_prune_groups_analysis(
            graph, pruned_nodes_clusterization, can_prune_analysis
        )
        can_prune_final_analysis = self._pruning_dimensions_analysis(
            graph, pruned_nodes_clusterization, can_prune_and_should_prune_analysis
        )
        self._filter_groups(pruned_nodes_clusterization, can_prune_final_analysis)
        return pruned_nodes_clusterization

    def _get_multiforward_nodes(self, graph: NNCFGraph) -> List[List[NNCFNode]]:
        """
        Groups nodes based on their `layer_name` property to determine groups of nodes belonging to
        a single weighted layer object in the model, i.e. the group of operations in the graph that reuse one and
        the same set of weights, and returns the groups that have more than one element.

        :return: List of lists of nodes; each list corresponds to a group of nodes united by the common
         underlying layer object of the original model.
        """
        ret = defaultdict(list)
        for node in graph.get_nodes_by_types(self._prune_operations_types):
            ret[node.layer_name].append(node)
        return [ret[module_identifier] for module_identifier in ret if len(ret[module_identifier]) > 1]

    def _pruning_dimensions_analysis(
        self,
        graph: NNCFGraph,
        pruned_nodes_clusterization: Clusterization,  # type: ignore[type-arg]
        can_prune_after_check: Dict[int, PruningAnalysisDecision],
    ) -> Dict[int, PruningAnalysisDecision]:
        """
        Checks:
        1) All nodes that were marked as prunable after the model analysis and compatibility check vs.
        pruning algo have a correct correspondent closing node on each path from self to outputs;
        2) Pruning dimensions of all nodes in all cluster groups are equal.

        :param graph: Graph to work with.
        :param pruned_nodes_clusterization: Pruned nodes clusterization.
        :param can_prune_after_check: Dict of node indices vs the decision made by previous steps;
            the decision is true only for the nodes that do not conflict with mask propagation and
            are supported by the NNCF pruning algorithm
        :return: Pruning node analysis after model analyzer, pruning algo compatibility and pruning dimensions checks.
        """

        nodes_of_group_with_non_eq_pruning_dim = self._check_internal_groups_dim(pruned_nodes_clusterization)
        can_prune_after_check_updated = can_prune_after_check.copy()
        for node_id, val in nodes_of_group_with_non_eq_pruning_dim.items():
            can_prune_after_check_updated[node_id] = can_prune_after_check_updated[node_id].join(val)

        return self._check_all_closing_nodes_are_feasible(graph, can_prune_after_check_updated)

    def _check_all_closing_nodes_are_feasible(
        self, graph: NNCFGraph, can_prune_after_check: Dict[int, PruningAnalysisDecision]
    ) -> Dict[int, PruningAnalysisDecision]:
        """
        Check all nodes that were marked as prunable after the model analysis and compatibility check vs.
        pruning algo have a correct correspondent closing node on each path from self to outputs.

        :param graph: Graph to work with.
        :param can_prune_after_check: Dict of node indices vs the decision made by previous steps;
            the decision is true only for the nodes that do not conflict with mask propagation and
            are supported by the NNCF pruning algorithm
        :return: Pruning node analysis after model analyzer, pruning algo compatibility and pruning dimensions checks.
        """
        mask_prop_algo = MaskPropagationAlgorithm(graph, self._pruning_operator_metatypes)
        can_prune_by_dim = mask_prop_algo.symbolic_mask_propagation(self._prune_operations_types, can_prune_after_check)

        can_prune_for_prunable_layers = {
            node_id: can_prune_after_check[node_id].join(can_prune_by_dim[node_id]) for node_id in can_prune_by_dim
        }

        can_prune_updated = can_prune_after_check.copy()
        can_prune_updated.update(can_prune_for_prunable_layers)
        return can_prune_updated

    def _check_internal_groups_dim(
        self, pruned_nodes_clusterization: Clusterization  # type: ignore[type-arg]
    ) -> Dict[int, PruningAnalysisDecision]:
        """
        Checks pruning dimensions of all nodes in each cluster group are equal and
        returns nodes of clusters that failed the check.

        :param pruned_nodes_clusterization: Pruned nodes clusterization.
        :returns: Pruning analysis decisions for nodes which have
            not equal pruning dimensions in a cluster they are belong to.
        """
        retval = {}
        for cluster in pruned_nodes_clusterization.get_all_clusters():
            has_equal_amount_of_channel = all(
                get_output_channels(cluster.elements[0]) == get_output_channels(node) for node in cluster.elements[1:]
            )
            if not has_equal_amount_of_channel:
                retval.update(
                    {
                        node.node_id: PruningAnalysisDecision(False, PruningAnalysisReason.INCOMPATIBLE_DIMS_IN_CLUSTER)
                        for node in cluster.elements
                    }
                )
        return retval

    def _should_prune_groups_analysis(
        self,
        graph: NNCFGraph,
        pruned_nodes_clusterization: Clusterization,  # type: ignore[type-arg]
        can_prune: Dict[int, PruningAnalysisDecision],
    ) -> Dict[int, PruningAnalysisDecision]:
        """
        Check whether all nodes in group can be pruned based on user-defined constraints and
        model analysis. Otherwise the whole group cannot be pruned.

        :param graph: Graph to work with.
        :param pruned_nodes_clusterization: Clusterization with pruning nodes groups.
        :param can_prune: Complete pruning analysis about each graph node.
        :return:
        """
        should_prune = {}
        for cluster in pruned_nodes_clusterization.get_all_clusters():
            should_prune_decisions = [self._is_module_prunable(graph, node) for node in cluster.elements]
            can_prune_decisions = [can_prune[node.node_id] for node in cluster.elements]
            decisions = [can.join(should) for can, should in zip(can_prune_decisions, should_prune_decisions)]
            if not all(decisions):
                updated_decisions = {}
                for node, decision in zip(cluster.elements, decisions):
                    if decision:
                        updated_decisions[node.node_id] = PruningAnalysisDecision(
                            False, PruningAnalysisReason.IN_GROUP_OF_UNPRUNABLE
                        )
                    else:
                        updated_decisions[node.node_id] = decision

                should_prune.update(updated_decisions)

        can_prune_updated = can_prune.copy()
        can_prune_updated.update(should_prune)
        return can_prune_updated

    def _filter_groups(
        self, pruned_nodes_clusterization: Clusterization, can_prune: Dict[int, PruningAnalysisDecision]  # type: ignore[type-arg]
    ) -> None:
        """
        Check whether all nodes in group can be pruned based on user-defined constraints and
        connections inside the network. Otherwise the whole group cannot be pruned and will be
        removed the clusterization.

        :param pruned_nodes_clusterization: Clusterization with pruning nodes groups.
        :param can_prune: Can this node be pruned or not.
        """
        for cluster in pruned_nodes_clusterization.get_all_clusters():
            nodes_decisions = [can_prune[node.node_id] for node in cluster.elements]
            nodes_names = [node.node_name for node in cluster.elements]
            if not all(nodes_decisions):
                cannot_prune_messages = []
                for name, decision in zip(nodes_names, nodes_decisions):
                    if not decision:
                        message = PruningAnalysisReason.message(name, decision)
                        if message:
                            cannot_prune_messages.append(message)

                nncf_logger.debug(
                    f'Could not prune node group [{", ".join(nodes_names)}], '
                    f'reason: {", ".join(cannot_prune_messages)}.'
                )
                pruned_nodes_clusterization.delete_cluster(cluster.id)
            else:
                nncf_logger.debug(f'Node group [{", ".join(nodes_names)}] will be pruned together.')

    def _is_module_prunable(self, graph: NNCFGraph, node: NNCFNode) -> PruningAnalysisDecision:
        """
        Check whether we should prune module corresponding to provided node
        according to algorithm parameters.

        :param graph: Graph to work with.
        :param node: Node to check.
        :return: Pruning analysis decision.
        """
        stop_propagation_ops = self._stop_propagation_op_metatype.get_all_op_aliases()
        types_to_track = self._prune_operations_types + stop_propagation_ops
        input_non_pruned_nodes = get_first_nodes_of_type(graph, types_to_track)
        node_name = node.node_name

        if not should_consider_scope(node_name, self._ignored_scopes, self._target_scopes):
            return PruningAnalysisDecision(False, PruningAnalysisReason.IGNORED_SCOPE)

        if not self._prune_first and node in input_non_pruned_nodes:
            return PruningAnalysisDecision(False, PruningAnalysisReason.FIRST_CONV)

        if is_grouped_conv(node) and not is_prunable_depthwise_conv(node):
            return PruningAnalysisDecision(False, PruningAnalysisReason.GROUP_CONV)

        if not self._prune_downsample_convs and is_conv_with_downsampling(node):
            return PruningAnalysisDecision(False, PruningAnalysisReason.DOWNSAMPLE_CONV)

        if is_batched_linear(node, graph):
            return PruningAnalysisDecision(False, PruningAnalysisReason.BATCHED_LINEAR)

        return PruningAnalysisDecision(True)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import dataclasses
import functools
import itertools
import operator
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

from nncf.common.factory import NNCFGraphFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.logging import nncf_logger
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import get_backend
from nncf.common.utils.timer import timer
from nncf.data.dataset import Dataset
from nncf.quantization.algorithms.accuracy_control.evaluator import Evaluator
from nncf.quantization.algorithms.accuracy_control.evaluator import MetricResults
from nncf.quantization.algorithms.accuracy_control.rank_functions import create_normalized_mse_func
from nncf.quantization.algorithms.accuracy_control.subset_selection import select_subset
from nncf.quantization.algorithms.pipeline import Pipeline
from nncf.quantization.algorithms.pipeline import collect_statistics

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")
CombinationKey = Tuple[int, ...]
Combination = Dict[str, Any]


def create_combinations(param_grid: Dict[str, List[Any]]) -> Dict[CombinationKey, Combination]:
    """
    Creates combinations as follows:
      * All keys in `param_grid` are numbered using integers from 0 to N = len(param_grid)-1
        Order of keys is used. Let key_j is a key from param_grid.keys() that corresponds
        integer j in {0, 1, ..., N}.
      * Set of combination keys (CK) are created as Cartesian product the following sets

          CK = {None, 0, 1, ..., num_val_0} x {None, 0, 1, ..., num_val_1} x ... x {None, 0, 1, ..., num_val_N},

        where num_val_j is a number of values in param_grid[key_j].
      * Creates combination for each combination key. If combination_key[i] is None then parameter with key_i
        name is not changed. Otherwise, the param_grid[key_i][combination_key[i]] value should be included
        to combination as new value for parameter with key_i name.

    :param param_grid: Dictionary with parameters names as keys and list of
        parameter settings to try as values.
    :return: Created combination.
    """
    simple_changes = []
    indices = []

    for param_name, values in param_grid.items():
        indices.append([None, *range(len(values))])
        simple_changes.append([{param_name: v} for v in values])

    combinations: Dict[CombinationKey, Combination] = {}

    for combination_key in itertools.product(*indices):
        combination: Combination = {}
        for param_idx, value_idx in enumerate(combination_key):
            if value_idx is None:
                continue
            combination.update(simple_changes[param_idx][value_idx])

        combinations[combination_key] = combination

    return combinations


def is_dataclass_instance(obj: Any) -> bool:
    """
    Returns `True` if object is a dataclass instance, `False` otherwise.

    :param obj: Object to check.
    :return: `True` if object is a dataclass instance, `False` otherwise.
    """
    return dataclasses.is_dataclass(obj) and not isinstance(obj, type)


def apply_combination(init_params: Dict[str, Any], combination: Combination) -> Dict[str, Any]:
    """
    Applies combination of parameters to initial parameters.

    :param init_params: Initial set of parameters.
    :param combination: Combination of parameters.
    :return: Returns `init_params` where some values of parameters were changed according to
        provided combination.
    """
    DELIMITER = ":"
    params = copy.deepcopy(init_params)
    for param_key, param_value in combination.items():
        if DELIMITER in param_key:
            main_key, *path_to_attr, attr_name = param_key.split(DELIMITER)
            obj = params[main_key]
            assert is_dataclass_instance(obj)
            for name in path_to_attr:
                obj = getattr(obj, name)
                assert is_dataclass_instance(obj)
            setattr(obj, attr_name, param_value)
        else:
            params[param_key] = param_value

    return params


def print_combination_and_score(
    title: str, combination: Combination, combination_score: Optional[float] = None
) -> None:
    """
    Prints combination and score.

    :param title: Title.
    :param combination: Combination to print.
    :param combination_score: Score of combination.
    """
    if not combination:
        message = "Parameters were not changed"
    else:
        message = ", ".join(f"{name} = {v}" for name, v in combination.items())
    message = f"{title} {message}"

    nncf_logger.info(message)

    if combination_score:
        nncf_logger.info(f"Score: {combination_score}")


def find_best_combination(
    combinations: Dict[CombinationKey, Combination],
    combination_score_func: Callable[[CombinationKey], float],
    param_grid: Dict[str, List[Any]],
) -> CombinationKey:
    """
    Finds best combination.

    :param combinations: Combinations.
    :param combination_score_func: Combination score function.
    :param param_grid: Dictionary with parameters names as keys and list of
        parameter settings to try as values.
    :return: Best combination key.
    """
    best_combination_key = tuple(None for _ in param_grid)
    best_combination_score = None

    for param_idx, (param_name, values) in enumerate(param_grid.items()):
        nncf_logger.info(f"Start search best value for the '{param_name}' parameter")
        values_indices = [None, *range(len(values))]
        param_best_combination_key = None
        param_best_combination_score = None

        for value_idx in values_indices:
            combination_key = (*best_combination_key[:param_idx], value_idx, *best_combination_key[param_idx + 1 :])
            combination_score = combination_score_func(combination_key)

            if param_best_combination_score is None or param_best_combination_score < combination_score:
                param_best_combination_score = combination_score
                param_best_combination_key = combination_key

            print_combination_and_score(
                "Current combination of parameters:", combinations[combination_key], combination_score
            )

        if best_combination_score is None or best_combination_score <= param_best_combination_score:
            best_combination_score = param_best_combination_score
            best_combination_key = param_best_combination_key

    print_combination_and_score(
        "Best combination of parameters:", combinations[best_combination_key], best_combination_score
    )

    return best_combination_key


class HyperparameterTuner:
    """
    This algorithm is used to find a best combination of parameters from `param_grid`.

    The `param_grid` in simple case is a dictionary with parameters names
    as keys and list of parameter settings to try as values.

        param_grid = {
            "param_name": [0.1, 0.2],
        }

    The parameters names should be same as in `pipeline_fn()` method.
    In case when "param_name" parameter is a dataclass object there is a way to specify settings
    to try for his fields using marker ":"

        param_grid = {
            "param_name:field_a": [10, 20],
            "param_name:field_b:x": [0.1, 0.2],
        }

    In the example above the `param_name` and "param_name:field_b" parameters are dataclasses.
    This rule is applied recursively.

    The algorithm works as follow: let we have the following `param_grid`

        param_grid = {
            "param_name_0" : [0.2, 0.4, 0.6],
            "param_name_1:x": [-1, -2, -3],
            "param_name_2": [True, False],
        }

    First of all, algorithm finds the best value for parameter "param_name_0".
    Further, taking into account the found value, the best value for the "param_name_1:x" parameter
    is sought. After that, taking into account the found values for "param_name_0" and "param_name_1:x"
    parameters, the best value for the "param_name_2" is sought.
    """

    def __init__(
        self,
        pipeline_fn: Callable[..., Pipeline],
        init_params: Dict[str, Any],
        param_grids: List[Dict[str, List[Any]]],
        calibration_dataset: Dataset,
        validation_fn: Callable[[Any, Iterable[Any]], Tuple[float, Union[None, List[float], List[List[TTensor]]]]],
        subset_size: int,
        initial_metric_results: MetricResults,
        quantized_metric_results: MetricResults,
    ):
        """
        :param pipeline_fn: Function to create pipeline.
        :param init_params: Initial set of parameters used to create algorithm.
        :param param_grid: Dictionary with parameters names as keys and list of
            parameter settings to try as values.
        :param calibration_dataset: Dataset used to collect statistics for algorithm.
        :param validation_fn: Validation function used to validated model.
        :param subset_size: Number of data items that should be selected
            from the dataset and used to validate model.
        :param initial_metric_results: Metric results for initial model.
        :param quantized_metric_results: Metric results for quantized with `init_params` model.
        """
        self._pipeline_fn = pipeline_fn
        self._init_params = init_params
        self._param_grids = param_grids
        self._calibration_dataset = calibration_dataset
        self._evaluator = Evaluator(validation_fn)
        self._subset_size = subset_size
        self._initial_metric_results = initial_metric_results
        self._quantized_metric_results = quantized_metric_results

        self._is_metric_mode = isinstance(self._initial_metric_results.values_for_each_item[0], float)

        # Will be initialized inside `apply()` method
        self._error_fn = None

        # Will be initialized inside `_prepare_pipeline_step()` method
        self._pipelines: Dict[CombinationKey, Pipeline] = {}
        self._step_index_to_statistics: Dict[int, StatisticPointsContainer] = {}

        self._calculated_scores: Dict[CombinationKey, float] = {}

    def apply(self, model: TModel, validation_dataset: Dataset) -> TModel:
        """
        Applies algorithm to input model.

        :param model: Input model.
        :param validation_dataset: Dataset used to validate resulted model.
        :return: Resulted model.
        """
        if self._is_metric_mode:
            self._error_fn = operator.sub
        else:
            self._error_fn = create_normalized_mse_func(get_backend(model))

        subset_indices = select_subset(
            self._subset_size,
            self._initial_metric_results.values_for_each_item,
            self._quantized_metric_results.values_for_each_item,
            self._error_fn,
        )

        step_model = model  # The model on which the `step_index`-th pipeline step will be executed
        best_settings = {}

        for step_index, step_param_grid in enumerate(self._param_grids):
            step_graph = NNCFGraphFactory.create(step_model)

            # If there are no parameters to optimize for the current step, simply execute
            # this pipeline step on the model.
            if not step_param_grid:
                # TODO(andrey-churkin): Think about how it can be avoided.
                params = apply_combination(self._init_params, best_settings)
                pipeline = self._pipeline_fn(**params)
                container = pipeline.get_statistic_points_for_step(step_index, step_model, step_graph)
                step_statistics = collect_statistics(container, step_model, step_graph, self._calibration_dataset)
                step_model = pipeline.run_step(step_index, step_statistics, step_model, step_graph)
                continue

            step_combinations = create_combinations(step_param_grid)

            nncf_logger.info(f"Start preparation for {step_index}-th pipeline step")
            with timer():
                self._prepare_pipeline_step(step_index, step_model, step_graph, step_combinations, best_settings)

            combination_score_fn = functools.partial(
                self._calculate_combination_score,
                step_index=step_index,
                step_model=step_model,
                step_graph=step_graph,
                dataset=validation_dataset,
                subset_indices=subset_indices,
            )

            nncf_logger.info("Start search best combination of parameters")
            with timer():
                step_best_combination_key = find_best_combination(
                    step_combinations, combination_score_fn, step_param_grid
                )

            best_settings.update(step_combinations[step_best_combination_key])
            pipeline = self._pipelines[step_best_combination_key]
            step_model = pipeline.run_step(
                step_index, self._step_index_to_statistics[step_index], step_model, step_graph
            )

        print_combination_and_score("Final best combination of parameters:", best_settings)

        return step_model

    def _prepare_pipeline_step(
        self,
        step_index: int,
        step_model: TModel,
        step_graph: NNCFGraph,
        step_combinations: Dict[CombinationKey, Combination],
        best_settings,
    ) -> None:
        """
        Creates a separate pipeline for each combination from step_combination.
        Each combination only changes the parameters of the `step_index`-th pipeline
        step. After that, combines the statistics required to execute the `step_index`-th
        pipeline step and collects them using `step_model`, `step_graph`, and the calibration
        dataset.

        :param step_index: Zero-based index of pipeline step that should be prepared.
        :param step_model: A model.
        :param step_graph: A graph assosiated with a model.
        :param step_combinations: Combinations that change parameters only for the step_index-th pipeline step.
        """
        # Create a separate pipeline for each combination

        # TODO(andrey-churkin): Think about how it can be avoided. In an ideal scenario,
        # we would have only one pipeline and set parameters directly within it.
        self._pipelines = {}
        for combination_key, combination in step_combinations.items():
            settings = {}
            settings.update(combination)
            settings.update(best_settings)
            kwargs = apply_combination(self._init_params, settings)
            self._pipelines[combination_key] = self._pipeline_fn(**kwargs)

        # Collect statistics required to execute `step_index`-th pipeline step
        containers = [
            pipeline.get_statistic_points_for_step(step_index, step_model, step_graph)
            for pipeline in self._pipelines.values()
        ]
        self._step_index_to_statistics[step_index] = collect_statistics(
            containers, step_model, step_graph, self._calibration_dataset
        )

    def _calculate_combination_score(
        self,
        combination_key: CombinationKey,
        step_index: int,
        step_model: TModel,
        step_graph: NNCFGraph,
        dataset: Dataset,
        subset_indices: List[int],
    ) -> float:
        """
        Calculates score for provided combination.

        :param combination_key: Combination key.
        :param initial_model: Input model.
        :param dataset: Dataset used to select data items for validation.
        :param subset_indices: Zero-based indices of data items that should be selected
            from the dataset and used to validate model.
        :return: Calculated score.
        """
        if combination_key in self._calculated_scores:
            return self._calculated_scores[combination_key]

        pipeline = self._pipelines[combination_key]
        model = pipeline.run_from_step(
            step_model, self._calibration_dataset, step_graph, step_index, self._step_index_to_statistics
        )

        score = self._validate_model(model, dataset, subset_indices)
        self._calculated_scores[combination_key] = score

        return score

    def _validate_model(self, model: TModel, dataset: Dataset, subset_indices: List[int]) -> float:
        """
        Validates input model on subset.

        :param model: Input model.
        :param dataset: Dataset used to select data items for validation.
        :param subset_indices: Zero-based indices of data items that should be selected
            from the dataset and used to validate model.
        :return: Calculated metric.
        """
        if self._is_metric_mode:
            metric_value, _ = self._evaluator.validate(model, dataset, subset_indices)
        else:
            approximate_outputs = self._evaluator.collect_values_for_each_item(model, dataset, subset_indices)
            reference_outputs = [self._initial_metric_results.values_for_each_item[i] for i in subset_indices]
            errors = [self._error_fn(a, b) for a, b in zip(reference_outputs, approximate_outputs)]
            metric_value = sum(errors) / len(errors)

        return metric_value
